@@ -260,7 +260,11 @@ RgC_Rg_add(GEN x, GEN y)
 {
   long k, lx = lg(x);
   GEN z = cgetg(lx, t_COL);
-  if (lx == 1) pari_err_TYPE2("+",x,y);
+  if (lx == 1)
+  {
+    if (isintzero(y)) return z;
+    pari_err_TYPE2("+",x,y);
+  }
   gel(z,1) = gadd(y,gel(x,1));
   for (k = 2; k < lx; k++) gel(z,k) = gcopy(gel(x,k));
   return z;
@@ -558,9 +562,9 @@ RgM_transmul(GEN x, GEN y)
   M = cgetg(ly, t_MAT);
   for (i=1; i<ly; i++)
   {
-    GEN xi = gel(x,i), c = cgetg(lx,t_COL);
+    GEN yi = gel(y,i), c = cgetg(lx,t_COL);
     gel(M,i) = c;
-    for (j=1; j<lx; j++) gel(c,j) = RgV_dotproduct_i(xi,gel(y,j),l);
+    for (j=1; j<lx; j++) gel(c,j) = RgV_dotproduct_i(yi,gel(x,j),l);
   }
   return M;
 }
