@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000-2004  The PARI group.
+/* Copyright (C) 2000-2004  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -11,7 +9,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /*******************************************************************/
 /*                                                                 */
@@ -44,17 +42,18 @@ typedef struct {
   GEN tozk;
   GEN topow;
   GEN topowden; /* topow x / topowden = basistoalg(x) */
+  GEN dn; /* NULL (we trust nf.zk) or a t_INT > 1 (an alg. integer has
+             denominator dividing dn, when expressed on nf.zk */
 } nflift_t;
 
 typedef struct
 {
   nflift_t *L;
   GEN nf;
-  GEN pol, polbase;
+  GEN pol, polbase; /* leading coeff is a t_INT */
   GEN fact;
   GEN pr;
   GEN Br, bound, ZC, BS_2;
-  GEN dn;
 } nfcmbf_t;
 
 /*******************************************************************/
@@ -87,7 +86,7 @@ FpM_ratlift(GEN M, GEN mod, GEN amax, GEN bmax, GEN denom)
   long i, j, h, l = lg(M);
   GEN a, N = cgetg_copy(M, &l);
   if (l == 1) return N;
-  h = lg(gel(M,1));
+  h = lgcols(M);
   for (j = 1; j < l; ++j)
   {
     gel(N,j) = cgetg(h, t_COL);
@@ -101,6 +100,20 @@ FpM_ratlift(GEN M, GEN mod, GEN amax, GEN bmax, GEN denom)
   return N;
 }
 GEN
+FpC_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
+{
+  pari_sp ltop = avma;
+  long j, l;
+  GEN a, Q = cgetg_copy(P, &l);
+  for (j = 1; j < l; ++j)
+  {
+    a = lift_to_frac(gel(P,j), mod, amax,bmax,denom);
+    if (!a) { avma = ltop; return NULL; }
+    gel(Q,j) = a;
+  }
+  return Q;
+}
+GEN
 FpX_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
 {
   pari_sp ltop = avma;
@@ -116,66 +129,6 @@ FpX_ratlift(GEN P, GEN mod, GEN amax, GEN bmax, GEN denom)
   return Q;
 }
 
-/* x, y, are ZYX, lc(y) is an integer, T is a ZY */
-static int
-ZXQX_dvd(GEN x, GEN y, GEN T)
-{
-  long dx, dy, dz, i, p, T_ismonic;
-  pari_sp av = avma, av2, lim;
-  GEN y_lead;
-
-  if (!signe(y)) pari_err(gdiver);
-  dy = degpol(y); y_lead = gel(y,dy+2);
-  if (typ(y_lead) == t_POL) y_lead = gel(y_lead, 2); /* t_INT */
-  /* if monic, no point in using pseudo-division */
-  if (gequal1(y_lead)) return signe(RgXQX_rem(x, y, T)) == 0;
-  T_ismonic = gequal1(leading_term(T));
-  dx = degpol(x);
-  if (dx < dy) return !signe(x);
-  (void)new_chunk(2);
-  x = RgX_recip_shallow(x)+2;
-  y = RgX_recip_shallow(y)+2;
-  /* pay attention to sparse divisors */
-  for (i = 1; i <= dy; i++)
-    if (!signe(gel(y,i))) gel(y,i) = NULL;
-  dz = dx-dy; p = dz+1;
-  av2 = avma; lim = stack_lim(av2,1);
-  for (;;)
-  {
-    GEN m, x0 = gel(x,0), y0 = y_lead, cx = content(x0);
-    x0 = gneg(x0); p--;
-    m = gcdii(cx, y0);
-    if (!equali1(m))
-    {
-      x0 = gdiv(x0, m);
-      y0 = diviiexact(y0, m);
-      if (equali1(y0)) y0 = NULL;
-    }
-    for (i=1; i<=dy; i++)
-    {
-      GEN c = gel(x,i); if (y0) c = gmul(y0, c);
-      if (gel(y,i)) c = gadd(c, gmul(x0,gel(y,i)));
-      if (typ(c) == t_POL) c = T_ismonic ? ZX_rem(c, T): RgX_rem(c, T);
-      gel(x,i) = c;
-    }
-    for (   ; i<=dx; i++)
-    {
-      GEN c = gel(x,i); if (y0) c = gmul(y0, c);
-      if (typ(c) == t_POL) c = T_ismonic ? ZX_rem(c, T): RgX_rem(c, T);
-      gel(x,i) = c;
-    }
-    do { x++; dx--; } while (dx >= 0 && !signe(gel(x,0)));
-    if (dx < dy) break;
-    if (low_stack(lim,stack_lim(av2,1)))
-    {
-      if(DEBUGMEM>1) pari_warn(warnmem,"ZXQX_dvd dx = %ld >= %ld",dx,dy);
-      gerepilecoeffs(av2,x,dx+1);
-    }
-  }
-  avma = av; return (dx < 0);
-}
-
-
 /*******************************************************************/
 /*              GCD in K[X], K NUMBER FIELD                        */
 /*******************************************************************/
@@ -201,10 +154,9 @@ GEN
 nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
 {
   pari_sp btop, st_lim, ltop = avma;
-  GEN lP, lQ, M, dsol, R, ax, bo, sol, mod = NULL;
+  GEN lP, lQ, M, dsol, R, bo, sol, mod = NULL;
   long vP = varn(P), vT = varn(T), dT = degpol(T), dM = 0, dR;
-  ulong p;
-  byteptr primepointer;
+  forprime_t S;
 
   if (!signe(P)) { if (Pnew) *Pnew = pol_0(vT); return gcopy(Q); }
   if (!signe(Q)) { if (Pnew) *Pnew = pol_1(vT);   return gcopy(P); }
@@ -215,11 +167,12 @@ nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
   if ( !((typ(lP)==t_INT && is_pm1(lP)) || (typ(lQ)==t_INT && is_pm1(lQ))) )
     den = mulii(den, gcdii(ZX_resultant(lP, T), ZX_resultant(lQ, T)));
 
+  init_modular(&S);
   btop = avma; st_lim = stack_lim(btop, 1);
-  primepointer = init_modular(&p);
-  for (;;)
+  for(;;)
   {
-    NEXT_PRIME_VIADIFF_CHECK(p, primepointer);
+    ulong p = u_forprime_next(&S);
+    if (!p) pari_err_OVERFLOW("nfgcd [ran out of primes]");
     /*Discard primes dividing disc(T) or lc(PQ) */
     if (!umodiu(den, p)) continue;
     if (DEBUGLEVEL>5) err_printf("nfgcd: p=%d\n",p);
@@ -231,19 +184,16 @@ nfgcd_all(GEN P, GEN Q, GEN T, GEN den, GEN *Pnew)
     if (dR == 0) { avma = ltop; if (Pnew) *Pnew = P; return pol_1(vP); }
     if (mod && dR > dM) continue; /* p divides Res(P/gcd, Q/gcd). Discard. */
 
-    R = RgXX_to_RgM(FlxX_to_ZXX(R), dT);
+    R = FlxX_to_Flm(R, dT);
     /* previous primes divided Res(P/gcd, Q/gcd)? Discard them. */
-    if (!mod || dR < dM) { M = R; mod = utoipos(p); dM = dR; continue; }
+    if (!mod || dR < dM) { M = ZM_init_CRT(R, p); mod = utoipos(p); dM = dR; continue; }
     if (low_stack(st_lim, stack_lim(btop, 1)))
     {
       if (DEBUGMEM>1) pari_warn(warnmem,"nfgcd");
       gerepileall(btop, 2, &M, &mod);
     }
 
-    ax = muliu(Fp_inv(utoipos(p), mod), p);
-    M = ZM_add(R, ZM_Z_mul(ZM_sub(M, R), ax));
-    mod = muliu(mod, p);
-    M = FpM_red(M, mod);
+    (void)ZM_incremental_CRT(&M,R, &mod,p);
     /* I suspect it must be better to take amax > bmax*/
     bo = sqrti(shifti(mod, -1));
     if ((sol = FpM_ratlift(M, mod, bo, bo, den)) == NULL) continue;
@@ -280,9 +230,8 @@ nffactormod(GEN nf, GEN x, GEN pr)
 
   nf = checknf(nf);
   vn = nf_get_varn(nf);
-  if (typ(x)!=t_POL) pari_err(typeer,"nffactormod");
-  if (varncmp(vx,vn) >= 0)
-    pari_err(talker,"polynomial variable must have highest priority in nffactormod");
+  if (typ(x)!=t_POL) pari_err_TYPE("nffactormod",x);
+  if (varncmp(vx,vn) >= 0) pari_err_PRIORITY("nffactormod", x, ">=", vn);
 
   modpr = nf_to_Fq_init(nf, &pr, &T, &p);
   xrd = nfX_to_FqX(x, nf, modpr);
@@ -319,8 +268,14 @@ static GEN
 RgX_int_normalize(GEN P)
 {
   GEN P0 = leading_term(P);
-  /* cater for t_POL (paranoia) */
-  if (typ(P0) == t_POL) P0 = gel(P0,2); /* non-0 constant */
+  /* cater for t_POL */
+  if (typ(P0) == t_POL)
+  {
+    P0 = gel(P0,2); /* non-0 constant */
+    P = shallowcopy(P);
+    gel(P,lg(P)-1) = P0; /* now leading term is a t_INT */
+  }
+  if (typ(P0) != t_INT) pari_err_BUG("RgX_int_normalize");
   if (is_pm1(P0)) return signe(P0) > 0? P: RgX_neg(P);
   return RgX_Rg_div(P, P0);
 }
@@ -341,7 +296,7 @@ get_den(GEN *pnf, GEN T)
     *pnf = nfinitall(T, nf_PARTIALFACT, DEFAULTPREC);
     D = nf_get_disc(proper_nf(*pnf));
     if (is_pm1(D)) return gen_1;
-    fa = Z_factor_limit(D, 0);
+    fa = absi_factor_limit(D, 0);
     P = gel(fa,1); q = gel(P, lg(P)-1);
     if (!BPSW_psp(q)) den = q; /* nf_get_disc(nf) may be incorrect */
   }
@@ -409,9 +364,9 @@ nfroots(GEN nf,GEN pol)
   if (!nf) return nfrootsQ(pol);
   T = get_nfpol(nf, &nf);
   RgX_check_ZX(T,"nfroots");
-  A = rnf_fix_pol(T,pol,1);
+  A = RgX_nffix("nfroots", T,pol,1);
   d = degpol(A);
-  if (d < 0) pari_err(zeropoler, "nfroots");
+  if (d < 0) pari_err_ROOTS0("nfroots");
   if (d == 0) return cgetg(1,t_VEC);
   if (d == 1)
   {
@@ -438,8 +393,8 @@ nfissplit(GEN nf, GEN x)
   pari_sp av = avma;
   long l;
   nf = checknf(nf);
-  x = rnf_fix_pol(nf_get_pol(nf), x, 1);
-  if (typ(x) != t_POL) pari_err(typeer, "nfissplit");
+  if (typ(x) != t_POL) pari_err_TYPE("nfissplit",x);
+  x = RgX_nffix("nfissplit", nf_get_pol(nf), x, 1);
   l = lg(nfsqff(nf, x, ROOTS_SPLIT, gen_1));
   avma = av; return l != 1;
 }
@@ -500,8 +455,7 @@ nf_pol_lift(GEN pol, GEN bound, nfcmbf_t *T)
   GEN t, x = cgetg(l,t_POL);
 
   x[1] = pol[1];
-  t = gel(pol,l-1);
-  gel(x,l-1) = mul_content(T->L->topowden, t);
+  gel(x,l-1) = mul_content(gel(pol,l-1), T->L->topowden);
   for (i=l-2; i>1; i--)
   {
     t = nf_bestlift_to_pol(gel(pol,i), bound, T->L);
@@ -540,14 +494,15 @@ fact_from_sqff(GEN rep, GEN A, GEN B, GEN y, GEN T, GEN bad)
     }
     else
     { /* compute valuations mod a prime of degree 1 (avoid coeff explosion) */
-      pari_sp av1 = avma;
-      long j;
       GEN quo, p, r, Bp, lb = leading_term(B), E = cgetalloc(t_VECSMALL,n+1);
-      byteptr pt = diffptr;
-      ulong pp = 0;
-      for (;; avma = av1)
+      pari_sp av1 = avma;
+      ulong pp;
+      long j;
+      forprime_t S;
+      u_forprime_init(&S, degpol(T), ULONG_MAX);
+      for (; ; avma = av1)
       {
-        NEXT_PRIME_VIADIFF_CHECK(pp, pt);
+        pp = u_forprime_next(&S);
         if (! umodiu(bad,pp) || !umodiu(lb, pp)) continue;
         p = utoipos(pp);
         r = FpX_oneroot(T, p);
@@ -596,11 +551,11 @@ nffactor(GEN nf,GEN pol)
   if (DEBUGLEVEL>2) { timer_start(&ti); err_printf("\nEntering nffactor:\n"); }
   T = get_nfpol(nf, &nf);
   RgX_check_ZX(T,"nffactor");
-  A = rnf_fix_pol(T,pol,1);
+  A = RgX_nffix("nffactor",T,pol,1);
   dA = degpol(A);
   if (dA <= 0) {
     avma = (pari_sp)(rep + 3);
-    return (dA == 0)? trivfact(): zerofact(varn(pol));
+    return (dA == 0)? trivial_fact(): zerofact(varn(pol));
   }
   A = Q_primpart( QXQX_normalize(A, T) );
   if (dA == 1) {
@@ -623,17 +578,18 @@ nffactor(GEN nf,GEN pol)
   return sort_factor_pol(rep, cmp_RgX);
 }
 
-/* assume x scalar or t_COL */
+/* assume x scalar or t_COL, G t_MAT */
 static GEN
 arch_for_T2(GEN G, GEN x)
 {
-  return (typ(x) == t_COL)? gmul(G,x): gmul(gel(G,1),x);
+  return (typ(x) == t_COL)? RgM_RgC_mul(G,x)
+                          : RgC_Rg_mul(gel(G,1),x);
 }
 static GEN
 arch_for_T2_prec(GEN G, GEN x, long prec)
 {
-  return (typ(x) == t_COL)? gmul(G, RgC_gtofp(x,prec))
-                          : gmul(gel(G,1), gtofp(x, prec));
+  return (typ(x) == t_COL)? RgM_RgC_mul(G, RgC_gtofp(x,prec))
+                          : RgC_Rg_mul(gel(G,1), gtofp(x, prec));
 }
 
 /* return a bound for T_2(P), P | polbase in C[X]
@@ -651,7 +607,7 @@ nf_Mignotte_bound(GEN nf, GEN polbase)
   long prec, i, j, d = degpol(polbase), n = nf_get_degree(nf), r1 = nf_get_r1(nf);
 
   binlS = bin = vecbinome(d-1);
-  if (!gequal1(lS)) binlS = gmul(lS, bin);
+  if (!isint1(lS)) binlS = ZC_Z_mul(bin,lS);
 
   N2 = cgetg(n+1, t_VEC);
   prec = gprecision(G);
@@ -664,26 +620,25 @@ nf_Mignotte_bound(GEN nf, GEN polbase)
     matGS = shallowtrans(matGS);
     for (j=1; j <= r1; j++) /* N2[j] = || sigma_j(S) ||_2 */
     {
-      gel(N2,j) = sqrtr( RgC_fpnorml2(gel(matGS,j), DEFAULTPREC) );
-      if (lg(N2[j]) < DEFAULTPREC) goto PRECPB;
+      GEN c = sqrtr( RgC_fpnorml2(gel(matGS,j), DEFAULTPREC) );
+      gel(N2,j) = c; if (!signe(c)) goto PRECPB;
     }
     for (   ; j <= n; j+=2)
     {
       GEN q1 = RgC_fpnorml2(gel(matGS,j  ), DEFAULTPREC);
       GEN q2 = RgC_fpnorml2(gel(matGS,j+1), DEFAULTPREC);
-      p1 = gmul2n(addrr(q1, q2), -1);
-      gel(N2,j) = gel(N2,j+1) = sqrtr(p1);
-      if (lg(N2[j]) < DEFAULTPREC) goto PRECPB;
+      GEN c = sqrtr( gmul2n(addrr(q1, q2), -1) );
+      gel(N2,j) = gel(N2,j+1) = c; if (!signe(c)) goto PRECPB;
     }
     if (j > n) break; /* done */
 PRECPB:
-    prec = (prec<<1)-2;
+    prec = precdbl(prec);
     remake_GM(nf, &F, prec); G = F.G;
     if (DEBUGLEVEL>1) pari_warn(warnprec, "nf_factor_bound", prec);
   }
 
   /* Take sup over 0 <= i <= d of
-   * sum_sigma | binom(d-1, i-1) ||sigma(S)||_2 + binom(d-1,i) lc(S) |^2 */
+   * sum_j | binom(d-1, i-1) ||sigma_j(S)||_2 + binom(d-1,i) lc(S) |^2 */
 
   /* i = 0: n lc(S)^2 */
   C = mului(n, sqri(lS));
@@ -692,7 +647,7 @@ PRECPB:
   for (i = 1; i < d; i++)
   {
     GEN B = gel(bin,i), L = gel(binlS,i+1);
-    GEN s = addri(mulir(B, gel(N2,1)),  L); /* j=1 */
+    GEN s = sqrr(addri(mulir(B, gel(N2,1)),  L)); /* j=1 */
     for (j = 2; j <= n; j++) s = addrr(s, sqrr(addri(mulir(B, gel(N2,j)), L)));
     if (mpcmp(C, s) < 0) C = s;
   }
@@ -702,38 +657,36 @@ PRECPB:
 /* return a bound for T_2(P), P | polbase
  * max |b_i|^2 <= 3^{3/2 + d} / (4 \pi d) [P]_2,
  * where [P]_2 is Bombieri's 2-norm
- * Sum over conjugates
-*/
+ * Sum over conjugates */
 static GEN
 nf_Beauzamy_bound(GEN nf, GEN polbase)
 {
   GEN lt, C, s, G = nf_get_G(nf), POL, bin;
-  long i,prec,precnf, d = degpol(polbase), n = nf_get_degree(nf);
-
-  precnf = gprecision(G);
-  prec   = MEDDEFAULTPREC;
+  long d = degpol(polbase), n = nf_get_degree(nf), prec   = MEDDEFAULTPREC;
   bin = vecbinome(d);
   POL = polbase + 2;
   /* compute [POL]_2 */
   for (;;)
   {
+    nffp_t F;
+    long i;
+
     s = real_0(prec);
     for (i=0; i<=d; i++)
     {
-      GEN p1 = gnorml2(arch_for_T2_prec(G, gel(POL,i), prec));
-      if (!signe(p1)) continue;
-      if (lg(p1) == 3) break;
+      GEN c = gel(POL,i);
+      if (gequal0(c)) continue;
+      c = gnorml2(arch_for_T2_prec(G, c, prec));
+      if (!signe(c)) goto PRECPB;
       /* s += T2(POL[i]) / binomial(d,i) */
-      s = addrr(s, gdiv(p1, gel(bin,i+1)));
+      s = addrr(s, divri(c, gel(bin,i+1)));
     }
-    if (i > d) break;
+    break;
 
-    prec = (prec<<1)-2;
-    if (prec > precnf)
-    {
-      nffp_t F; remake_GM(nf, &F, prec); G = F.G;
-      if (DEBUGLEVEL>1) pari_warn(warnprec, "nf_factor_bound", prec);
-    }
+PRECPB:
+    prec = precdbl(prec);
+    remake_GM(nf, &F, prec); G = F.G;
+    if (DEBUGLEVEL>1) pari_warn(warnprec, "nf_factor_bound", prec);
   }
   lt = leading_term(polbase);
   s = mulri(s, muliu(sqri(lt), n));
@@ -766,12 +719,12 @@ nf_root_bounds(GEN P, GEN T)
   T = get_nfpol(T, &nf);
 
   P = Q_primpart(P);
-  prec = ZXY_max_lg(P) + 1;
+  prec = ZXX_max_lg(P) + 1;
   l = lg(P);
   if (nf && nf_get_prec(nf) >= prec)
     R = nf_get_roots(nf);
   else
-    R = cleanroots(T, prec);
+    R = QX_complex_roots(T, prec);
   lR = lg(R);
   V = cgetg(lR, t_VEC);
   Ps = cgetg(l, t_POL); /* sigma (P) */
@@ -795,7 +748,7 @@ L2_bound(GEN nf, GEN den)
   GEN M, L, prep, T = nf_get_pol(nf), tozk = nf_get_invzk(nf);
   long prec;
 
-  prec = ZX_max_lg(T) + ZM_max_lg(tozk) - 2;
+  prec = nbits2prec(bit_accuracy(ZX_max_lg(T)) + bit_accuracy(ZM_max_lg(tozk)));
   (void)initgaloisborne(nf, den, prec, &L, &prep, NULL);
   M = vandermondeinverse(L, RgX_gtofp(T,prec), den, prep);
   return RgM_fpnorml2(RgM_mul(tozk,M), DEFAULTPREC);
@@ -891,7 +844,7 @@ init_trace(trace_data *T, GEN S, nflift_t *L, GEN q)
 
   T->dPinvS = ZM_mul(L->iprk, S);
   l = lg(S);
-  h = lg(T->dPinvS[1]);
+  h = lgcols(T->dPinvS);
   T->PinvSdbl = (double**)cgetg(l, t_MAT);
   init_dalloc();
   for (j = 1; j < l; j++)
@@ -913,9 +866,9 @@ static void
 update_trace(trace_data *T, long k, long i)
 {
   if (!T) return;
-  T->S1[k]       = T->S1[i];
-  T->dPinvS[k]   = T->dPinvS[i];
-  T->PinvSdbl[k] = T->PinvSdbl[i];
+  gel(T->S1,k)     = gel(T->S1,i);
+  gel(T->dPinvS,k) = gel(T->dPinvS,i);
+  T->PinvSdbl[k]   = T->PinvSdbl[i];
 }
 
 /* reduce coeffs mod (T,pk), then center mod pk */
@@ -939,30 +892,36 @@ FqX_centermod(GEN z, GEN T, GEN pk, GEN pks2)
 }
 
 typedef struct {
-  GEN lt, C, C2lt, C2ltpol;
+  GEN lt, C, Clt, C2lt, C2ltpol;
 } div_data;
 
 static void
-init_div_data(div_data *D, GEN pol, GEN C)
+init_div_data(div_data *D, GEN pol, nflift_t *L)
 {
-  GEN C2lt, lc = leading_term(pol), lt = is_pm1(lc)? NULL: absi(lc);
+  GEN C = mul_content(L->topowden, L->dn);
+  GEN C2lt, Clt, lc = leading_term(pol), lt = is_pm1(lc)? NULL: absi(lc);
   if (C)
   {
     GEN C2 = sqri(C);
-    C2lt = lt ? mulii(C2, lt): C2;
+    if (lt) {
+      C2lt = mulii(C2, lt);
+      Clt = mulii(C,lt);
+    } else {
+      C2lt = C2;
+      Clt = C;
+    }
   }
   else
-    C2lt = lt;
+    C2lt = Clt = lt;
   D->lt = lt;
   D->C = C;
+  D->Clt = Clt;
   D->C2lt = C2lt;
   D->C2ltpol = C2lt? RgX_Rg_mul(pol, C2lt): pol;
 }
 static void
 update_target(div_data *D, GEN pol)
-{
-  D->C2ltpol = D->C? RgX_Rg_mul(pol, D->C): pol;
-}
+{ D->C2ltpol = D->Clt? RgX_Rg_mul(pol, D->Clt): pol; }
 
 /* nb = number of modular factors; return a "good" K such that naive
  * recombination of up to maxK modular factors is not too costly */
@@ -983,15 +942,15 @@ cmbf_maxK(long nb)
 static GEN
 nfcmbf(nfcmbf_t *T, long klim, long *pmaxK, int *done)
 {
-  GEN pol = T->pol, nf = T->nf, famod = T->fact, dn = T->dn, bound = T->bound;
-  GEN lt, ltdn, nfpol = nf_get_pol(nf);
+  GEN nf = T->nf, famod = T->fact, bound = T->bound;
+  GEN ltdn, nfpol = nf_get_pol(nf);
   long K = 1, cnt = 1, i,j,k, curdeg, lfamod = lg(famod)-1, dnf = degpol(nfpol);
   pari_sp av0 = avma;
   GEN Tpk = T->L->Tpk, pk = T->L->pk, pks2 = shifti(pk,-1);
   GEN ind      = cgetg(lfamod+1, t_VECSMALL);
   GEN deg      = cgetg(lfamod+1, t_VECSMALL);
   GEN degsofar = cgetg(lfamod+1, t_VECSMALL);
-  GEN fa       = cgetg(lfamod+1, t_COL);
+  GEN fa       = cgetg(lfamod+1, t_VEC);
   const double Bhigh = get_Bhigh(lfamod, dnf);
   trace_data _T1, _T2, *T1, *T2;
   div_data D;
@@ -1000,17 +959,13 @@ nfcmbf(nfcmbf_t *T, long klim, long *pmaxK, int *done)
   timer_start(&ti);
 
   *pmaxK = cmbf_maxK(lfamod);
-  init_div_data(&D, pol, T->L->topowden);
-  ltdn = mul_content(D.lt, dn);
-  lt = D.lt; /* to be restored at the end */
+  init_div_data(&D, T->pol, T->L);
+  ltdn = mul_content(D.lt, T->L->dn);
   {
     GEN q = ceil_safe(sqrtr(T->BS_2));
-    GEN t1,t2, lt2dn;
+    GEN t1,t2, lt2dn = mul_content(ltdn, D.lt);
     GEN trace1   = cgetg(lfamod+1, t_MAT);
     GEN trace2   = cgetg(lfamod+1, t_MAT);
-
-    lt2dn= mul_content(ltdn, D.lt);
-
     for (i=1; i <= lfamod; i++)
     {
       pari_sp av = avma;
@@ -1060,9 +1015,8 @@ nextK:
       pari_sp av;
 
       av = avma;
-      /* d - 1 test */
       if (T1)
-      {
+      { /* d-1 test */
         t = get_trace(ind, T1);
         if (rtodbl(RgC_fpnorml2(t,DEFAULTPREC)) > Bhigh)
         {
@@ -1070,9 +1024,8 @@ nextK:
           avma = av; goto NEXT;
         }
       }
-      /* d - 2 test */
       if (T2)
-      {
+      { /* d-2 test */
         t = get_trace(ind, T2);
         if (rtodbl(RgC_fpnorml2(t,DEFAULTPREC)) > Bhigh)
         {
@@ -1088,37 +1041,34 @@ nextK:
         if (y) q = gmul(y, q);
         y = FqX_centermod(q, Tpk, pk, pks2);
       }
-      /* y = C*lt*\prod_{i in ind} famod[i] */
       y = nf_pol_lift(y, bound, T);
       if (!y)
       {
         if (DEBUGLEVEL>3) err_printf("@");
         avma = av; goto NEXT;
       }
-      /* y is apparently in O_K[X], in fact in (Z[Y]/nf.pol)[X] due to the
-       * multiplication by C. Try out this candidate factor */
+      /* y = topowden*dn*lt*\prod_{i in ind} famod[i] is apparently in O_K[X],
+       * in fact in (Z[Y]/nf.pol)[X] due to multiplication by C = topowden*dn.
+       * Try out this candidate factor */
       q = RgXQX_divrem(D.C2ltpol, y, nfpol, ONLY_DIVIDES);
       if (!q)
       {
         if (DEBUGLEVEL>3) err_printf("*");
         avma = av; goto NEXT;
       }
-      /* pol in O_K[X] with leading coeff lt in Z,
+      /* Original T->pol in O_K[X] with leading coeff lt in Z,
        * y = C*lt \prod famod[i] is in O_K[X] with leading coeff in Z
-       * q = C^2 * lt * pol / y = C * (lt * pol) / (lt*\prod famod[i]), a
-       * K-rational factor, in fact in Z[Y]/nf.pol)[X] as above. */
-
-      /* found a factor */
-      if (D.C2lt) y = RgX_int_normalize(y); /* monic */
-      gel(fa,cnt++) = y;
-      /* fix up pol */
-      pol = q;
+       * q = C^2*lt*pol / y = C * (lt*pol) / (lt*\prod famod[i]) is a
+       * K-rational factor, in fact in Z[Y]/nf.pol)[X] as above, with
+       * leading term C*lt. */
+      update_target(&D, q);
+      gel(fa,cnt++) = D.C2lt? RgX_int_normalize(y): y; /* make monic */
       for (i=j=k=1; i <= lfamod; i++)
       { /* remove used factors */
         if (j <= K && i == ind[j]) j++;
         else
         {
-          famod[k] = famod[i];
+          gel(famod,k) = gel(famod,i);
           update_trace(T1, k, i);
           update_trace(T2, k, i);
           deg[k] = deg[i]; k++;
@@ -1128,8 +1078,6 @@ nextK:
       *pmaxK = cmbf_maxK(lfamod);
       if (lfamod < 2*K) goto END;
       i = 1; curdeg = deg[ind[1]];
-
-      update_target(&D, pol);
       if (DEBUGLEVEL > 2)
       {
         err_printf("\n"); timer_printf(&ti, "to find factor %Ps",y);
@@ -1151,16 +1099,17 @@ NEXT:
   }
 END:
   *done = 1;
-  if (degpol(pol) > 0)
+  if (degpol(D.C2ltpol) > 0)
   { /* leftover factor */
-    if (D.C2lt) pol = RgX_int_normalize(pol);
+    GEN q = D.C2ltpol;
+    if (D.C2lt) q = RgX_int_normalize(q);
     if (lfamod >= 2*K)
     { /* restore leading coefficient [#930] */
-      if (lt) pol = RgX_Rg_mul(pol, lt);
+      if (D.lt) q = RgX_Rg_mul(q, D.lt);
       *done = 0; /* ... may still be reducible */
     }
     setlg(famod, lfamod+1);
-    gel(fa,cnt++) = pol;
+    gel(fa,cnt++) = q;
   }
   if (DEBUGLEVEL>6) err_printf("\n");
   if (cnt == 2) {
@@ -1189,8 +1138,8 @@ nf_chk_factors(nfcmbf_t *T, GEN P, GEN M_L, GEN famod, GEN pk)
   if (DEBUGLEVEL>3) err_printf("special_pivot output:\n%Ps\n",piv);
 
   r  = lg(piv)-1;
-  list = cgetg(r+1, t_COL);
-  init_div_data(&D, pol, T->L->topowden);
+  list = cgetg(r+1, t_VEC);
+  init_div_data(&D, pol, T->L);
   for (i = 1;;)
   {
     pari_sp av = avma;
@@ -1242,7 +1191,7 @@ ZqX(GEN P, GEN pk, GEN T, GEN proj)
 static GEN
 ZqX_normalize(GEN P, GEN lt, nflift_t *L)
 {
-  GEN R = lt? gmul(Fp_inv(lt, L->pk), P): P;
+  GEN R = lt? RgX_Rg_mul(P, Fp_inv(lt, L->pk)): P;
   return ZqX(R, L->pk, L->Tpk, L->ZqProj);
 }
 
@@ -1264,13 +1213,13 @@ static GEN
 get_R(GEN M)
 {
   GEN R;
-  long i, l, prec = DEFAULTPREC + divsBIL( gexpo(M) );
+  long i, l, prec = nbits2prec( gexpo(M) + 64 );
 
   for(;;)
   {
-    R = Q_from_QR(M, prec);
+    R = gaussred_from_QR(M, prec);
     if (R) break;
-    prec = (prec-1)<<1;
+    prec = precdbl(prec);
   }
   l = lg(R);
   for (i=1; i<l; i++) gcoeff(R,i,i) = gen_1;
@@ -1308,7 +1257,7 @@ max_radius(GEN PRK, GEN B)
   pari_sp av = avma;
   long i, j, d = lg(PRK)-1;
 
-  S = RgM_inv( get_R(PRK) ); if (!S) pari_err(precer,"max_radius");
+  S = RgM_inv( get_R(PRK) ); if (!S) pari_err_PREC("max_radius");
   for (i=1; i<=d; i++)
   {
     GEN s = gen_0;
@@ -1382,7 +1331,7 @@ nf_LLL_cmbf(nfcmbf_t *T, long rec)
 {
   const double BitPerFactor = 0.4; /* nb bits / modular factor */
   nflift_t *L = T->L;
-  GEN famod = T->fact, ZC = T->ZC, Br = T->Br, P = T->pol, dn = T->dn;
+  GEN famod = T->fact, ZC = T->ZC, Br = T->Br, P = T->pol, dn = T->L->dn;
   long dnf = nf_get_degree(T->nf), dP = degpol(P);
   long i, C, tmax, n0;
   GEN lP, Bnorm, Tra, T2, TT, CM_L, m, list, ZERO, Btra;
@@ -1560,14 +1509,15 @@ nf_combine_factors(nfcmbf_t *T, GEN polred, long klim)
 }
 
 static GEN
-nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, GEN ltdn, GEN init_fa, long nbf,
+nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, GEN init_fa, long nbf,
               long fl, nflift_t *L)
 {
-  GEN z, Cltdnx_r, C2ltdnpol, C = L->topowden;
-  GEN Cltdn  = mul_content(C, ltdn); /* t_INT */
-  GEN C2ltdn = mul_content(C,Cltdn); /* t_INT */
+  GEN z, Cltx_r, ltdn;
   long i, m;
+  div_data D;
 
+  init_div_data(&D, pol, L);
+  ltdn = mul_content(D.lt, L->dn);
   if (L->Tpk)
   {
     int cof = (degpol(pol) > nbf); /* non trivial cofactor ? */
@@ -1578,18 +1528,20 @@ nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, GEN ltdn, GEN init_fa, long nbf,
   }
   else
     z = rootpadicfast(polred, L->p, L->k);
-  Cltdnx_r = deg1pol_shallow(Cltdn? Cltdn: gen_1, NULL, varn(pol));
-  C2ltdnpol  = C2ltdn? RgX_Rg_mul(pol, C2ltdn): pol;
+  Cltx_r = deg1pol_shallow(D.Clt? D.Clt: gen_1, NULL, varn(pol));
   for (m=1,i=1; i<lg(z); i++)
   {
     GEN q, r = gel(z,i);
-
+    pari_sp av;
+    /* lt*dn*topowden * r = Clt * r */
     r = nf_bestlift_to_pol(ltdn? gmul(ltdn,r): r, NULL, L);
-    gel(Cltdnx_r,2) = gneg(r); /* check P(r) == 0 */
-    q = RgXQX_divrem(C2ltdnpol, Cltdnx_r, nfpol, ONLY_DIVIDES); /* integral */
+    av = avma;
+    gel(Cltx_r,2) = gneg(r); /* check P(r) == 0 */
+    q = RgXQX_divrem(D.C2ltpol, Cltx_r, nfpol, ONLY_DIVIDES); /* integral */
+    avma = av;
+    /* don't go on with q, usually much larger that C2ltpol */
     if (q) {
-      C2ltdnpol = C2ltdn? RgX_Rg_mul(q, Cltdn): q;
-      if (Cltdn) r = gdiv(r, Cltdn);
+      if (D.Clt) r = gdiv(r, D.Clt);
       gel(z,m++) = r;
     }
     else if (fl == ROOTS_SPLIT) return cgetg(1, t_VEC);
@@ -1600,10 +1552,10 @@ nf_DDF_roots(GEN pol, GEN polred, GEN nfpol, GEN ltdn, GEN init_fa, long nbf,
 
 /* returns a factor of T in Fp of degree <= maxf, NULL if none exist */
 static GEN
-get_good_factor(GEN T, GEN p, long maxf)
+get_good_factor(GEN T, ulong p, long maxf)
 {
   pari_sp av = avma;
-  GEN r, list = gel(FpX_factor(T,p), 1);
+  GEN r, list = gel(Flx_factor(ZX_to_Flx(T,p),p), 1);
   if (maxf == 1)
   { /* deg.1 factors are best */
     r = gel(list,1);
@@ -1649,8 +1601,8 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
 {
   GEN nfpol = nf_get_pol(nf), bad = mulii(nf_get_disc(nf), nf_get_index(nf));
   long maxf, nfdeg = degpol(nfpol), dpol = degpol(polbase), nbf = 0;
-  byteptr pt = diffptr;
-  ulong pp = 0;
+  ulong pp;
+  forprime_t S;
   pari_timer ti_pr;
 
   if (DEBUGLEVEL>3) timer_start(&ti_pr);
@@ -1661,8 +1613,9 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
   *Tp = NULL;
 
   maxf = get_maxf(nfdeg);
+  (void)u_forprime_init(&S, 2, ULONG_MAX);
   /* select pr such that pol has the smallest number of factors, ct attempts */
-  for (;;)
+  while ((pp = u_forprime_next(&S)))
   {
     GEN aT, apr, ap, amodpr, red, r, fa = NULL;
     long anbf;
@@ -1670,14 +1623,13 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
     pari_sp av2 = avma;
 
     /* first step : select prime of high inertia degree */
-    NEXT_PRIME_VIADIFF_CHECK(pp, pt);
     if (! umodiu(bad,pp)) continue;
     if (*lt) { ltp = umodiu(*lt, pp); if (!ltp) continue; }
-    ap = utoipos(pp);
-    r = get_good_factor(nfpol, ap, maxf);
+    r = get_good_factor(nfpol, pp, maxf);
     if (!r) continue;
 
-    apr = primedec_apply_kummer(nf,r,1,ap);
+    ap = utoipos(pp);
+    apr = primedec_apply_kummer(nf, Flx_to_ZX(r), 1, ap);
     amodpr = zk_to_Fq_init(nf,&apr,&aT,&ap);
 
     /* second step : evaluate factorisation mod apr */
@@ -1691,12 +1643,10 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
     }
     else
     {
-      GEN q;
       if (ltp) red = FqX_normalize(red, aT,ap);
       if (!FqX_is_squarefree(red,aT,ap)) { avma = av2; continue; }
-      q = powiu(ap, degpol(aT));
-      anbf = fl == FACTORS? FqX_split_by_degree(&fa, red, q, aT, ap)
-                          : FqX_split_deg1(&fa, red, q, aT, ap);
+      anbf = fl == FACTORS? FqX_split_by_degree(&fa, red, aT, ap)
+                          : FqX_split_deg1(&fa, red, aT, ap);
     }
     if (fl == ROOTS_SPLIT && anbf < dpol) return anbf;
     if (anbf <= 1)
@@ -1717,8 +1667,10 @@ nf_pick_prime(long ct, GEN nf, GEN polbase, long fl,
     if (DEBUGLEVEL>3)
       err_printf("%3ld %s at prime\n  %Ps\nTime: %ld\n",
                  anbf, fl == FACTORS?"factors": "roots", apr, timer_delay(&ti_pr));
-    if (--ct <= 0) return nbf;
+    if (--ct <= 0) break;
   }
+  if (!nbf) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
+  return nbf;
 }
 
 /* assume lt(T) is a t_INT and T square free */
@@ -1726,7 +1678,7 @@ static GEN
 nfsqff_trager(GEN u, GEN T, GEN dent)
 {
   long k = 0, i, lx;
-  GEN P, x0, fa, n = ZX_ZXY_rnfequation(T, u, &k);
+  GEN U, P, x0, mx0, fa, n = ZX_ZXY_rnfequation(T, u, &k);
   int tmonic;
   if (DEBUGLEVEL>4) err_printf("nfsqff_trager: choosing k = %ld\n",k);
 
@@ -1737,13 +1689,16 @@ nfsqff_trager(GEN u, GEN T, GEN dent)
   tmonic = is_pm1(leading_term(T));
   P = cgetg(lx,t_COL);
   x0 = deg1pol_shallow(stoi(-k), gen_0, varn(T));
+  mx0 = deg1pol_shallow(stoi(k), gen_0, varn(T));
+  U = RgXQX_translate(u, mx0, T);
+  if (!tmonic) U = Q_primpart(U);
   for (i=lx-1; i>0; i--)
   {
-    GEN f = gel(fa,i), F = RgXQX_translate(f, x0, T);
-    if (!tmonic) F = Q_primpart(F);
-    F = nfgcd(u, F, T, dent);
+    GEN f = gel(fa,i), F = nfgcd(U, f, T, dent);
+    F = RgXQX_translate(F, x0, T);
+    /* F = gcd(f, u(t - x0)) [t + x0] = gcd(f(t + x0), u), more efficient */
     if (typ(F) != t_POL || degpol(F) == 0)
-      pari_err(talker,"reducible modulus in factornf");
+      pari_err_IRREDPOL("factornf [modulus]",T);
     gel(P,i) = QXQX_normalize(F, T);
   }
   return P;
@@ -1758,15 +1713,16 @@ polfnf(GEN a, GEN T)
   long dA;
   int tmonic;
 
-  if (typ(a)!=t_POL || typ(T)!=t_POL) pari_err(typeer,"polfnf");
+  if (typ(a)!=t_POL) pari_err_TYPE("polfnf",a);
+  if (typ(T)!=t_POL) pari_err_TYPE("polfnf",T);
   T = Q_primpart(T); tmonic = is_pm1(leading_term(T));
   RgX_check_ZX(T,"polfnf");
-  A = Q_primpart( QXQX_normalize(rnf_fix_pol(T,a,1), T) );
+  A = Q_primpart( QXQX_normalize(RgX_nffix("polfnf",T,a,1), T) );
   dA = degpol(A);
   if (dA <= 0)
   {
     avma = (pari_sp)(rep + 3);
-    return (dA == 0)? trivfact(): zerofact(varn(A));
+    return (dA == 0)? trivial_fact(): zerofact(varn(A));
   }
   bad = dent = ZX_disc(T);
   if (tmonic) dent = indexpartial(T, dent);
@@ -1805,7 +1761,10 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
   if (DEBUGLEVEL>2) { timer_start(&ti); timer_start(&ti_tot); }
   n = degpol(nfpol);
   /* deg = 1 => irreducible */
-  if (dpol == 1) return mkvec(QXQX_normalize(pol, nfpol));
+  if (dpol == 1) {
+    if (fl == FACTORS) return mkvec(QXQX_normalize(pol, nfpol));
+    return mkvec(gneg(gdiv(gel(pol,2),gel(pol,3))));
+  }
   if (typ(nf)==t_POL || nfsqff_use_Trager(n,dpol))
   {
     GEN z;
@@ -1842,8 +1801,9 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
   }
   L.tozk = nf_get_invzk(nf);
   L.topow= Q_remove_denom(nf_get_zk(nf), &L.topowden);
+  if (is_pm1(den)) den = NULL;
+  L.dn = den;
   T.ZC = L2_bound(nf, den);
-  T.dn = is_pm1(den)? NULL: den;
   T.Br = nf_root_bounds(pol, nf); if (lt) T.Br = gmul(T.Br, lt);
 
   /* C0 = bound for T_2(Q_i), Q | P */
@@ -1869,8 +1829,7 @@ nfsqff(GEN nf, GEN pol, long fl, GEN den)
   polred = ZqX_normalize(polbase, lt, &L); /* monic */
 
   if (fl != FACTORS) {
-    GEN z = nf_DDF_roots(pol, polred, nfpol, mul_content(lt, T.dn),
-                         init_fa, nbf, fl, &L);
+    GEN z = nf_DDF_roots(pol, polred, nfpol, init_fa, nbf, fl, &L);
     if (lg(z) == 1) return cgetg(1, t_VEC);
     return z;
   }
@@ -1907,7 +1866,7 @@ nfroots_split(GEN nf, GEN pol)
 {
   GEN T = get_nfpol(nf,&nf), den = get_den(&nf, T);
   pari_sp av = avma;
-  GEN z = gerepilecopy(av, nfsqff(nf, pol, 2, den));
+  GEN z = gerepilecopy(av, nfsqff(nf, pol, ROOTS_SPLIT, den));
   return (lg(z) == 1)? NULL: mkvec2(z, nf);
 }
 
@@ -1947,20 +1906,20 @@ nf_pick_prime_for_units(GEN nf, prklift_t *P)
 {
   GEN nfpol = nf_get_pol(nf), bad = mulii(nf_get_disc(nf), nf_get_index(nf));
   GEN aT, amodpr, apr, ap = NULL, r = NULL;
-  long maxf, nfdeg = degpol(nfpol);
-  byteptr pt = diffptr;
-  ulong pp = 0;
+  long nfdeg = degpol(nfpol), maxf = get_maxf(nfdeg);
+  ulong pp;
+  forprime_t S;
 
-  maxf = get_maxf(nfdeg);
-  for (;;)
+  (void)u_forprime_init(&S, 2, ULONG_MAX);
+  while ( (pp = u_forprime_next(&S)) )
   {
-    NEXT_PRIME_VIADIFF_CHECK(pp, pt);
     if (! umodiu(bad,pp)) continue;
-    ap = utoipos(pp);
-    r = get_good_factor(nfpol, ap, maxf);
+    r = get_good_factor(nfpol, pp, maxf);
     if (r) break;
   }
-  apr = primedec_apply_kummer(nf,r,1,ap);
+  if (!r) pari_err_OVERFLOW("nf_pick_prime [ran out of primes]");
+  ap = utoipos(pp);
+  apr = primedec_apply_kummer(nf, Flx_to_ZX(r), 1, ap);
   amodpr = zk_to_Fq_init(nf,&apr,&aT,&ap);
   P->pr = apr;
   P->q = pr_norm(apr);
@@ -1997,14 +1956,14 @@ nfcyclo_root(GEN nf, long n_cyclo, prklift_t *P)
   GEN z, nfpol = nf_get_pol(nf), pol = polcyclo(n_cyclo, MAXVARN);
   long nbf, deg = degpol(pol); /* = eulerphi(n_cyclo) */
   if (P->L->Tp)
-    nbf = FqX_split_deg1(&init_fa, pol, P->q, P->L->Tp, P->L->p);
+    nbf = FqX_split_deg1(&init_fa, pol, P->L->Tp, P->L->p);
   else
   {
     ulong p = itou(P->L->p);
     nbf = Flx_nbroots(ZX_to_Flx(pol,p), p);
   }
   if (nbf != deg) return NULL; /* no roots in residue field */
-  z = nf_DDF_roots(pol, pol, nfpol, gen_1, init_fa, nbf, ROOTS_SPLIT, P->L);
+  z = nf_DDF_roots(pol, pol, nfpol, init_fa, nbf, ROOTS_SPLIT, P->L);
   if (lg(z) == 1) { avma = av; return NULL; } /* no roots */
   return gel(z,1);
 }
@@ -2015,21 +1974,23 @@ nfcyclo_root(GEN nf, long n_cyclo, prklift_t *P)
 static long
 guess_roots(GEN nf)
 {
-  long c = 0, nfdegree = nf_get_degree(nf), l;
+  long c = 0, nfdegree = nf_get_degree(nf), B = nfdegree + 20, l;
   ulong p = 2;
-  byteptr pt = diffptr+1;
-  GEN T = nf_get_pol(nf), nbroots = NULL;
-  pari_sp av = avma;
+  GEN T = nf_get_pol(nf), D = nf_get_disc(nf), index = nf_get_index(nf);
+  GEN nbroots = NULL;
+  forprime_t S;
+  pari_sp av;
 
-  /* result must be stationnary (counter c) for at least nfdegree+20 loops */
-  for (l=1;; l++)
+  (void)u_forprime_init(&S, 3, ULONG_MAX);
+  av = avma;
+  /* result must be stationnary (counter c) for at least B loops */
+  for (l=1; (p = u_forprime_next(&S)); l++)
   {
-    GEN old, Tp, F, pf_1;
+    GEN old, F, pf_1, Tp;
     long i, nb, gcdf = 0;
 
-    NEXT_PRIME_VIADIFF_CHECK(p,pt);
-    Tp = ZX_to_Flx(T,p);
-    if (!Flx_is_squarefree(Tp, p)) continue;
+    if (!umodiu(D,p) || !umodiu(index,p)) continue;
+    Tp = ZX_to_Flx(T,p); /* squarefree */
     F = Flx_nbfact_by_degree(Tp, &nb, p);
     /* the gcd of the p^f - 1 is p^(gcd of the f's) - 1 */
     for (i = 1; i <= nfdegree; i++)
@@ -2044,14 +2005,56 @@ guess_roots(GEN nf)
       err_printf("p=%ld; gcf(f(P/p))=%ld; nbroots | %Ps",p, gcdf, nbroots);
     /* if same result go on else reset the stop counter [c] */
     if (old && equalii(nbroots,old))
-    { if (!is_bigint(nbroots) && ++c > nfdegree + 20) break; }
+    { if (!is_bigint(nbroots) && ++c > B) break; }
     else
       c = 0;
   }
+  if (!nbroots) pari_err_OVERFLOW("guess_roots [ran out of primes]");
   if (DEBUGLEVEL>5) err_printf("%ld loops\n",l);
   avma = av; return itos(nbroots);
 }
 
+/* T(x) an irreducible ZX. Is it of the form Phi_n(c \pm x) ?
+ * Return NULL if not, and a root of 1 of maximal order in Z[x]/(T) otherwise
+ *
+ * N.B. Set n_squarefree = 1 if n is squarefree, and 0 otherwise.
+ * This last parameter is inconvenient, but it allows a cheap
+ * stringent test. (n guessed from guess_roots())*/
+static GEN
+ZXirred_is_cyclo_translate(GEN T, long n_squarefree)
+{
+  long r, m, d = degpol(T);
+  GEN T1, c = divis_rem(gel(T, d+1), d, &r); /* d-1 th coeff divisible by d ? */
+  /* The trace coefficient of polcyclo(n) is \pm1 if n is square free, and 0
+   * otherwise. */
+  if (!n_squarefree)
+  { if (r) return NULL; }
+  else
+  {
+    if (r < -1)
+    {
+      r += d;
+      c = subiu(c, 1);
+    }
+    else if (r == d-1)
+    {
+      r = -1;
+      c = addiu(c, 1);
+    }
+    if (r != 1 && r != -1) return NULL;
+  }
+  if (signe(c)) /* presumably Phi_guess(c \pm x) */
+    T = RgX_translate(T, negi(c));
+  if (!n_squarefree) T = RgX_deflate_max(T, &m);
+  /* presumably Phi_core(guess)(\pm x), cyclotomic iff original T was */
+  T1 = ZX_graeffe(T);
+  if (ZX_equal(T, T1)) /* T = Phi_n, n odd */
+    return deg1pol_shallow(gen_m1, negi(c), varn(T));
+  else if (ZX_equal(T1, ZX_unscale(T, gen_m1))) /* T = Phi_{2n}, nodd */
+    return deg1pol_shallow(gen_1, c, varn(T));
+  return NULL;
+}
+
 static GEN
 trivroots(void) { return mkvec2(gen_2, gen_m1); }
 /* Number of roots of unity in number field [nf]. */
@@ -2085,24 +2088,40 @@ rootsof1(GEN nf)
     long p = LP[i];
     /* Cheap test: can Q(zeta_{2p}) be a subset of K ? */
     if (p == 2)
-    { /* check that 2 | n and v_p(disc K) >= n/2 */
-      if (!odd(nfdegree) && vali(disc) >= nfdegree / 2) continue;
+    { /* 2 | n and v_p(disc K) >= n/2 ? */
       if (LE[i] == 1) continue;
+      if (!odd(nfdegree) && vali(disc) >= nfdegree / 2) continue;
     }
     else
-    { /* check that p-1 | n and v_p(disc K) >= (p-2) n/(p-1) */
-      long v = Z_lval(disc, p);
-      if (nfdegree % (p-1) == 0 && v >= nfdegree/(p-1) * (p-2)) continue;
+    { /* p-1 | n and v_p(disc K) >= (p-2) n/(p-1) ? */
+      ulong r, q = udivuu_rem(nfdegree, p-1, &r);
+      if (r == 0 && (ulong)Z_lval(disc, p) >= q * (p-2)) continue;
     }
     nbguessed /= upowuu(p, LE[i]);
     LE[i] = 0;
   }
   if (DEBUGLEVEL>2)
-    timer_printf(&ti, "adding ramification conditions [guess = %ld]", nbguessed);
+    timer_printf(&ti, "after ramification conditions [guess = %ld]", nbguessed);
   if (nbguessed == 2) return trivroots();
+  av = avma;
+
+  /* Step 1.5 : test if nf.pol == subst(polcyclo(nbguessed), x, \pm x+c) */
+  if (eulerphiu_fact(fa) == (ulong)nfdegree)
+  {
+    GEN elt = ZXirred_is_cyclo_translate(nf_get_pol(nf),
+                                         uissquarefree_fact(fa));
+    if (elt)
+    {
+      if (DEBUGLEVEL>2)
+        timer_printf(&ti, "checking for cyclotomic polynomial [yes]");
+      return gerepilecopy(av, mkvec2(utoipos(nbguessed), elt));
+    }
+    avma = av;
+  }
+  if (DEBUGLEVEL>2)
+    timer_printf(&ti, "checking for cyclotomic polynomial [no]");
 
   /* Step 2 : choose a prime ideal for local lifting */
-  av = avma;
   P.L = &L; nf_pick_prime_for_units(nf, &P);
   if (DEBUGLEVEL>2)
     timer_printf(&ti, "choosing prime %Ps, degree %ld",
@@ -2114,6 +2133,7 @@ rootsof1(GEN nf)
   /* lift and reduce pr^k */
   if (DEBUGLEVEL>2) err_printf("Lift pr^k; GSmin wanted: %Ps\n",C0);
   bestlift_init((long)mybestlift_bound(C0), nf, P.pr, C0, P.L);
+  P.L->dn = NULL;
   if (DEBUGLEVEL>2) timer_start(&ti);
 
   /* Step 4 : actual computation of roots */
@@ -2181,11 +2201,11 @@ rootsof1_kannan(GEN nf)
       y = fincke_pohst(mkvec(R), utoipos(N), N * N, 0, NULL);
       if (y) break;
     }
-    prec = (prec<<1)-2;
+    prec = precdbl(prec);
     if (DEBUGLEVEL) pari_warn(warnprec,"rootsof1",prec);
     nf = nfnewprec_shallow(nf,prec);
   }
-  if (itos(ground(gel(y,2))) != N) pari_err(bugparier,"rootsof1 (bug1)");
+  if (itos(ground(gel(y,2))) != N) pari_err_BUG("rootsof1 (bug1)");
   w = gel(y,1); ws = itos(w);
   if (ws == 2) { avma = av; return trivroots(); }
 
@@ -2195,6 +2215,6 @@ rootsof1_kannan(GEN nf)
     z = is_primitive_root(nf, d, gel(list,i), ws);
     if (z) return gerepilecopy(av, mkvec2(w, z));
   }
-  pari_err(bugparier,"rootsof1");
+  pari_err_BUG("rootsof1");
   return NULL; /* not reached */
 }
