@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -11,7 +9,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 #include "pari.h"
 #include "paripriv.h"
@@ -138,7 +136,7 @@ millerrabin(GEN n, long k)
   long i;
   MR_Jaeschke_t S;
 
-  if (typ(n) != t_INT) pari_err(arither1);
+  if (typ(n) != t_INT) pari_err_TYPE("millerrabin",n);
   if (signe(n)<=0) return 0;
   /* If |n| <= 3, check if n = +- 1 */
   if (lgefint(n)==3 && (ulong)(n[2])<=3) return (n[2] != 1);
@@ -319,8 +317,8 @@ u_LucasMod(ulong n, ulong P, ulong N)
   return v;
 }
 
-static int
-u_IsLucasPsP(ulong n)
+int
+uislucaspsp(ulong n)
 {
   long i, v;
   ulong b, z, m2, m = n + 1;
@@ -394,11 +392,53 @@ uu_coprime(ulong n, ulong u)
   return gcduodd(n, u) == 1;
 }
 
-/* Fl_BPSW_psp */
+/* composite strong 2-pseudoprime < 1016801 whose prime divisors are > 101 */
+static int
+is_2_prp_101(ulong n)
+{
+  switch(n) {
+  case 42799:
+  case 49141:
+  case 88357:
+  case 90751:
+  case 104653:
+  case 130561:
+  case 196093:
+  case 220729:
+  case 253241:
+  case 256999:
+  case 271951:
+  case 280601:
+  case 357761:
+  case 390937:
+  case 458989:
+  case 486737:
+  case 489997:
+  case 514447:
+  case 580337:
+  case 741751:
+  case 838861:
+  case 873181:
+  case 877099:
+  case 916327:
+  case 976873:
+  case 983401: return 1;
+  } return 0;
+}
+
+static int
+u_2_prp(ulong n)
+{
+  Fl_MR_Jaeschke_t S;
+  Fl_init_MR_Jaeschke(&S, n);
+  return Fl_bad_for_base(&S, 2) == 0;
+}
+static int
+uBPSW_psp(ulong n) { return (u_2_prp(n) && uislucaspsp(n)); }
+
 int
 uisprime(ulong n)
 {
-  Fl_MR_Jaeschke_t S;
   if (n < 103)
     switch(n)
     {
@@ -430,7 +470,7 @@ uisprime(ulong n)
       case 101: return 1;
       default: return 0;
     }
-  if (!(n & 1)) return 0;
+  if (!odd(n)) return 0;
 #ifdef LONG_IS_64BIT
   /* 16294579238595022365 = 3*5*7*11*13*17*19*23*29*31*37*41*43*47*53
    *  7145393598349078859 = 59*61*67*71*73*79*83*89*97*101 */
@@ -447,51 +487,22 @@ uisprime(ulong n)
       !uu_coprime(n, 4269855901UL)) return 0;
 #endif
   if (n < 10427) return 1;
-  Fl_init_MR_Jaeschke(&S, n);
-  if (Fl_bad_for_base(&S, 2)) return 0;
-  if (n < 1016801) switch(n) {
-  /* strong 2-pseudoprimes without prime divisors < 103. All have 2 prime
-   * divisor, one of which is <= 661 */
-    case 42799:
-    case 49141:
-    case 88357:
-    case 90751:
-    case 104653:
-    case 130561:
-    case 196093:
-    case 220729:
-    case 253241:
-    case 256999:
-    case 271951:
-    case 280601:
-    case 357761:
-    case 390937:
-    case 458989:
-    case 486737:
-    case 489997:
-    case 514447:
-    case 580337:
-    case 741751:
-    case 838861:
-    case 873181:
-    case 877099:
-    case 916327:
-    case 976873:
-    case 983401: return 0;
-    default: return 1;
-  }
-  return u_IsLucasPsP(n);
+  if (n < 1016801) return !is_2_prp_101(n) && u_2_prp(n);
+  return uBPSW_psp(n);
+}
+
+/* assume no prime divisor <= 101 */
+int
+uisprime_101(ulong n)
+{
+  if (n < 10427) return 1;
+  if (n < 1016801) return !is_2_prp_101(n) && u_2_prp(n);
+  return uBPSW_psp(n);
 }
 
 /* assume no prime divisor <= 661 */
 int
-uisprime_nosmalldiv(ulong n)
-{
-  Fl_MR_Jaeschke_t S;
-  Fl_init_MR_Jaeschke(&S, n);
-  if (Fl_bad_for_base(&S, 2)) return 0;
-  return u_IsLucasPsP(n);
-}
+uisprime_661(ulong n) { return uBPSW_psp(n); }
 
 long
 BPSW_psp(GEN N)
@@ -500,7 +511,7 @@ BPSW_psp(GEN N)
   MR_Jaeschke_t S;
   int k;
 
-  if (typ(N) != t_INT) pari_err(arither1);
+  if (typ(N) != t_INT) pari_err_TYPE("BPSW_psp",N);
   if (signe(N) <= 0) return 0;
   if (lgefint(N) == 3) return uisprime((ulong)N[2]);
   if (!mod2(N)) return 0;
@@ -526,16 +537,40 @@ BPSW_psp(GEN N)
   avma = av; return k;
 }
 
-/* no prime divisor <= 661 */
+/* can we write n = x^k ? Assume N has no prime divisor <= 2^14.
+ * Not memory clean */
+long
+isanypower_nosmalldiv(GEN N, GEN *px)
+{
+  GEN x = N, y;
+  ulong mask = 7;
+  long ex, k = 1;
+  forprime_t T;
+  while (Z_issquareall(x, &y)) { k <<= 1; x = y; }
+  while ( (ex = is_357_power(x, &y, &mask)) ) { k *= ex; x = y; }
+  (void)u_forprime_init(&T, 11, ULONG_MAX);
+  /* stop when x^(1/k) < 2^14 */
+  while ( (ex = is_pth_power(x, &y, &T, 15)) ) { k *= ex; x = y; }
+  *px = x; return k;
+}
+
+/* no prime divisor <= 2^14 (> 661) */
 long
 BPSW_psp_nosmalldiv(GEN N)
 {
   pari_sp av;
   MR_Jaeschke_t S;
+  long l = lgefint(N);
   int k;
 
-  if (lgefint(N) == 3) return uisprime_nosmalldiv((ulong)N[2]);
+  if (l == 3) return uisprime_661((ulong)N[2]);
   av = avma;
+  /* N large: test for pure power, rarely succeeds, but requires < 1% of
+   * compositeness test times */
+  if (bit_accuracy(l) > 512 && isanypower_nosmalldiv(N, &N) != 1)
+  {
+    avma = av; return 0;
+  }
   init_MR_Jaeschke(&S, N);
   k = (!bad_for_base(&S, gen_2) && IsLucasPsP(N));
   avma = av; return k;
@@ -562,20 +597,25 @@ BPSW_isprime_small(GEN x)
 }
 
 /* Brillhart, Lehmer, Selfridge test (Crandall & Pomerance, Th 4.1.5)
- * N^(1/3) <= f fully factored, f | N-1 */
+ * N^(1/3) <= f fully factored, f | N-1. If pl831(p) is true for
+ * any prime divisor p of f, then any divisor of N is 1 mod f.
+ * In that case return 1 iff N is prime */
 static int
 BLS_test(GEN N, GEN f)
 {
   GEN c1, c2, r, q;
-  if (cmpii(sqri(f), N) < 0) return 1; /* Pocklington case */
   q = dvmdii(N, f, &r);
   if (!is_pm1(r)) return 0;
   c2 = dvmdii(q, f, &c1);
-  if (Z_issquare(subii(c2, shifti(c1,2)))) return 0;
-  return 1;
+  /* N = 1 + f c1 + f^2 c2, 0 <= c_i < f; check whether it is of the form
+   * (1 + f a)(1 + fb) */
+  return ! Z_issquare(subii(sqri(c1), shifti(c2,2)));
 }
 
-/*assume n>=2*/
+/*assume N>1, p^e || N-1. Find a witness a(p) such that
+ * a^(N-1) = 1 (mod N)
+ * a^(N-1)/p - 1 invertible mod N.
+ * Proves that any divisor of N is 1 mod p^e */
 static ulong
 pl831(GEN N, GEN p)
 {
@@ -592,100 +632,135 @@ pl831(GEN N, GEN p)
     if (!equalii(g,N)) return 0;
   }
 }
+/* Assume x BPSW pseudoprime. Return gen_0 if not prime, and a primality
+ * certificate otherwise */
+static GEN isprimePL(GEN N);
+static GEN
+check_prime(GEN p)
+{
+  if (BPSW_isprime_small(p)) return gen_1;
+  if (expi(p) <= 250) return isprimePL(p);
+  return isprimeAPRCL(p)? gen_2: gen_0;
+}
+/* initialize Selfridge / Pocklington-Lehmer test, return 0
+ * if proven composite */
+static int
+selfridge_init(GEN N, GEN *pF, GEN *pf)
+{
+  GEN cbrtN = sqrtnint(N, 3);
+  GEN N_1 = addis(N,-1);
+  GEN F = Z_factor_until(N_1, sqri(cbrtN));
+  GEN f = factorback(F); /* factored part of N-1, f^3 > N */
+  *pF = gel(F,1);
+  *pf = f;
+  /* smooth or N^(1/2)-smooth (Pocklington) or N^(1/3)-smooth (BLS) */
+  return equalii(f, N_1) || cmpii(sqri(f), N) > 0 || BLS_test(N,f);
+}
 /* Assume N is a strong BPSW pseudoprime, Pocklington-Lehmer primality proof.
  *
- * flag 0: return gen_1 (prime), gen_0 (composite)
- * flag 1: return gen_0 (composite), gen_1 (small prime), matrix (large prime)
+ * return gen_0 (non-prime), gen_1 (small prime), matrix (large prime)
  *
  * The matrix has 3 columns, [a,b,c] with
  * a[i] prime factor of N-1,
  * b[i] witness for a[i] as in pl831
  * c[i] isprimePL(a[i]) */
 static GEN
-isprimePL(GEN N, long flag)
+isprimePL(GEN N)
 {
   pari_sp ltop = avma;
-  long i, l, t = typ(N);
+  long i, l;
   int eps;
-  GEN C, F = NULL;
+  GEN C, P, W, R, F, f;
 
-  if (t == t_VEC)
-  { /* [ N, [p1,...,pk] ], (p_i) list of pseudoprime divisors of N-1 */
-    F = gel(N,2);
-    N = gel(N,1); t = typ(N);
-  }
-  if (t != t_INT) pari_err(arither1);
+  if (typ(N) != t_INT) pari_err_TYPE("isprimePL",N);
   eps = cmpis(N,2);
-  if (eps<=0) return eps? gen_0: gen_1;
-
+  if (eps <= 0) return eps? gen_0: gen_1;
+  /* N > 2 */
+  if (!selfridge_init(N, &F, &f)) { avma = ltop; return gen_0; }
   if (DEBUGLEVEL>3)
-    err_printf("Pocklington-Lehmer: proving primality of N = %Ps\n", N);
-  N = absi(N);
-  if (!F)
   {
-    GEN cbrtN = gsqrtn(N, utoi(3), NULL, nbits2prec((expi(N)+2)/3));
-    GEN N_1 = addis(N,-1), f;
-    F = Z_factor_until(N_1, sqri(floorr(cbrtN)));
-    f = factorback(F); F = gel(F,1);
-    if (!equalii(f, N_1) && !BLS_test(N,f)) {
-      if (DEBUGLEVEL>3)
-        err_printf("Pocklington-Lehmer: N-1 not smooth enough --> Failure. Factored up to %Ps (%.3Ps%%)\n", f, divri(itor(f,3), N));
-      avma = ltop; return gen_0;
-    }
-    if (DEBUGLEVEL>3)
-      err_printf("Pocklington-Lehmer: N-1 factored up to %Ps! (%.3Ps%%)\n", f, divri(itor(f,3), N));
-  }
-  if (DEBUGLEVEL>3)
+    err_printf("Pocklington-Lehmer: proving primality of N = %Ps\n", N);
+    err_printf("Pocklington-Lehmer: N-1 factored up to %Ps! (%.3Ps%%)\n", f, divri(itor(f,LOWDEFAULTPREC), N));
     err_printf("Pocklington-Lehmer: N-1 smooth enough! Computing certificate\n");
+  }
   C = cgetg(4,t_MAT); l = lg(F);
-  gel(C,1) = cgetg(l,t_COL);
-  gel(C,2) = cgetg(l,t_COL);
-  gel(C,3) = cgetg(l,t_COL);
+  gel(C,1) = P = cgetg(l,t_COL);
+  gel(C,2) = W = cgetg(l,t_COL);
+  gel(C,3) = R = cgetg(l,t_COL);
   for(i=1; i<l; i++)
   {
-    GEN p = gel(F,i), r;
+    GEN p = gel(F,i);
     ulong witness = pl831(N,p);
 
     if (!witness) { avma = ltop; return gen_0; }
-    gmael(C,1,i) = icopy(p);
-    gmael(C,2,i) = utoi(witness);
-    if (DEBUGLEVEL>3)
-      err_printf("Pocklington-Lehmer: recursively proving primality of p = %Ps\n", p);
-    if (!flag) r = BPSW_isprime(p)? gen_1: gen_0;
-    else
-    {
-      if (BPSW_isprime_small(p)) r = gen_1;
-      else if (expi(p) > 250)   r = isprimeAPRCL(p)? gen_2: gen_0;
-      else                      r = isprimePL(p,flag);
+    gel(P,i) = icopy(p);
+    gel(W,i) = utoipos(witness);
+    gel(R,i) = check_prime(p);
+    if (gel(R,i) == gen_0)
+    { /* composite in prime factorisation ! */
+      err_printf("Not a prime: %Ps", p);
+      pari_err_BUG("isprimePL [false prime number]");
     }
-    gmael(C,3,i) = r;
-    if (r == gen_0) pari_err(talker,"False prime number %Ps in isprimePL", p);
   }
-  if (!flag) { avma = ltop; return gen_1; }
   return gerepileupto(ltop,C);
 }
-static long
-isprimeSelfridge(GEN x) { return (isprimePL(x,0)==gen_1); }
 
-/* assume x a BPSW pseudoprime */
+/* F is a vector of BPSW pseudoprimes of N. For each of then, find a PL witness,
+ * then prove primality. If (avoid_last) do not apply test to last entry */
+static long
+isprimeSelfridge(GEN N, GEN F, int avoid_last)
+{
+  long i, l = lg(F);
+  if (avoid_last) l--;
+  for(i = 1; i < l; i++)
+  {
+    GEN p = gel(F,i);
+    if (! pl831(N, p)) return 0;
+    if (DEBUGLEVEL>3)
+      err_printf("Pocklington-Lehmer: recursively proving primality of p = %Ps\n", p);
+    if (!BPSW_isprime(p)) return 0;
+  }
+  return 1;
+}
+
+/* assume N a BPSW pseudoprime, in particular, it is odd > 2 */
 long
 BPSW_isprime(GEN N)
 {
   pari_sp av = avma;
   long l, res;
-  GEN fa, P, F, p, N_1;
+  ulong B;
+  GEN fa, P, E, p, U, F, N_1;
 
   if (BPSW_isprime_small(N)) return 1;
-  N_1 = subis(N,1); fa = Z_factor_limit(N_1, minss(1L<<19, maxprime()));
-  l = lg(gel(fa,1))-1; p = gcoeff(fa,l,1);
-  F = diviiexact(N_1,  powii(p, gcoeff(fa,l,2)));
+  N_1 = subis(N,1);
+  B = minuu(1UL<<19, maxprime());
+  fa = Z_factor_limit(N_1, B);
   P = gel(fa,1);
+  E = gel(fa,2);
+  l = lg(P)-1;
+  p = gel(P,l);
+  U = powii(p, gel(E,l)); /* (possibly) unfactored part of N-1 */
+  if (cmpiu(U, B) <= 0) { U = gen_1; F = N_1; }
+  else
+  { /* l >= 2, since 2 and U divide  N-1 */
+    if (l == 2)
+      F = powii(gel(P,1), gel(E,1));
+    else
+      F = diviiexact(N_1,  U);
+  }
   /* N-1 = F U, F factored, U possibly composite */
-  if (cmpii(powiu(F, 3), N) >= 0) /* half-smooth, F >= N^(1/3) */
-    res = BLS_test(N, F)
-          && isprimeSelfridge(mkvec2(N,vecslice(P,1,l-1)));
-  else if (BPSW_psp_nosmalldiv(p)) /* smooth */
-    res = isprimeSelfridge(mkvec2(N,P));
+  if (U == gen_1) /* smooth */
+    res = isprimeSelfridge(N, P, 0);
+  else if (cmpii(F, U) >= 0) /* 1/2-smooth */
+    res = isprimeSelfridge(N, P, 1);
+  else if (cmpii(sqri(F), U) >= 0) /* 1/3-smooth */
+  {
+    res = BLS_test(N, F);
+    if (res) res = isprimeSelfridge(N, P, 1);
+  }
+  else if (BPSW_psp_nosmalldiv(p)) /* smooth after all */
+    res = isprimeSelfridge(N, P, 0);
   else
     res = isprimeAPRCL(N);
   avma = av; return res;
@@ -697,10 +772,10 @@ gisprime(GEN x, long flag)
   switch (flag)
   {
     case 0: return map_proto_lG(isprime,x);
-    case 1: return map_proto_GL(isprimePL,x,1);
+    case 1: return map_proto_G(isprimePL,x);
     case 2: return map_proto_lG(isprimeAPRCL,x);
   }
-  pari_err(flagerr,"gisprime");
+  pari_err_FLAG("gisprime");
   return NULL;
 }
 
@@ -713,110 +788,459 @@ isprime(GEN x) { return BPSW_psp(x) && BPSW_isprime(x); }
 /**                                                                   **/
 /***********************************************************************/
 
-/* assume all primes up to 59359 are precomputed */
-ulong
-uprime(long n)
-{
-  byteptr p;
-  ulong prime;
+static struct {
+  ulong p;
+  long n;
+} prime_table[] = {
+  {           0,          0},
+  {        7919,       1000},
+  {       17389,       2000},
+  {       27449,       3000},
+  {       37813,       4000},
+  {       48611,       5000},
+  {       59359,       6000},
+  {       70657,       7000},
+  {       81799,       8000},
+  {       93179,       9000},
+  {      104729,      10000},
+  {      224737,      20000},
+  {      350377,      30000},
+  {      479909,      40000},
+  {      611953,      50000},
+  {      746773,      60000},
+  {      882377,      70000},
+  {     1020379,      80000},
+  {     1159523,      90000},
+  {     1299709,     100000},
+  {     2750159,     200000},
+  {     7368787,     500000},
+  {    15485863,    1000000},
+  {    32452843,    2000000},
+  {    86028121,    5000000},
+  {   179424673,   10000000},
+  {   373587883,   20000000},
+  {   982451653,   50000000},
+  {  2038074743,  100000000},
+  {  4000000483UL,189961831},
+  {  4222234741UL,200000000},
+#if BITS_IN_LONG == 64
+  { 11037271757,  500000000},
+  { 22801763489, 1000000000},
+  { 47055833459, 2000000000},
+  {122430513841, 5000000000},
+  {200000000507, 8007105083},
+#endif
+};
+static const int prime_table_len = sizeof(prime_table)/sizeof(prime_table[0]);
 
-  if (n <= 0) pari_err(talker, "n-th prime meaningless if n = %ld",n);
-  if (n < 1000) {
-    p = diffptr;
-    prime = 0;
-  } else if (n < 2000) {
-    n -= 1000; p = diffptr+1000;
-    prime = 7919;
-  } else if (n < 3000) {
-    n -= 2000; p = diffptr+2000;
-    prime = 17389;
-  } else if (n < 4000) {
-    n -= 3000; p = diffptr+3000;
-    prime = 27449;
-  } else if (n < 5000) {
-    n -= 4000; p = diffptr+4000;
-    prime = 37813;
-  } else if (n < 6000) {
-    n -= 5000; p = diffptr+5000;
-    prime = 48611;
-  } else if (n < 10000 || maxprime() < 500000) {
-    n -= 6000; p = diffptr+6000;
-    prime = 59359;
-  } else if (n < 20000) {
-    n -= 10000; p = diffptr+10000;
-    prime = 104729;
-  } else if (n < 30000) {
-    n -= 20000; p = diffptr+20000;
-    prime = 224737;
-  } else if (n < 40000) {
-    n -= 30000; p = diffptr+30000;
-    prime = 350377;
-  } else {
-    n -= 40000; p = diffptr+40000;
-    prime = 479909;
+/* find prime closest to n in prime_table. */
+static long
+prime_table_closest_p(ulong n)
+{
+  long i;
+  for (i = 1; i < prime_table_len; i++)
+  {
+    ulong p = prime_table[i].p;
+    if (p > n)
+    {
+      ulong u = n - prime_table[i-1].p;
+      if (p - n > u) i--;
+      break;
+    }
   }
-  while (n--) NEXT_PRIME_VIADIFF_CHECK(prime,p);
-  return prime;
+  if (i == prime_table_len) i = prime_table_len - 1;
+  return i;
+}
+
+/* return the n-th successor of prime p > 2 */
+static GEN
+prime_successor(ulong p, ulong n)
+{
+  forprime_t S;
+  ulong i;
+  forprime_init(&S, utoipos(p+1), NULL);
+  for (i = 1; i < n; i++) (void)forprime_next(&S);
+  return forprime_next(&S);
+}
+/* find the N-th prime */
+static GEN
+prime_table_find_n(ulong N)
+{
+  byteptr d;
+  ulong n, p, maxp = maxprime();
+  long i;
+  for (i = 1; i < prime_table_len; i++)
+  {
+    n = prime_table[i].n;
+    if (n > N)
+    {
+      ulong u = N - prime_table[i-1].n;
+      if (n - N > u) i--;
+      break;
+    }
+  }
+  if (i == prime_table_len) i = prime_table_len - 1;
+  p = prime_table[i].p;
+  n = prime_table[i].n;
+  if (n > N && p > maxp)
+  {
+    i--;
+    p = prime_table[i].p;
+    n = prime_table[i].n;
+  }
+  /* if beyond prime table, then n <= N */
+  d = diffptr + n;
+  if (n > N)
+  {
+    n -= N;
+    do { n--; PREC_PRIME_VIADIFF(p,d); } while (n) ;
+  }
+  else if (n < N)
+  {
+    n = N-n;
+    if (p > maxp) return prime_successor(p, n);
+    do {
+      if (!*d) return prime_successor(p, n);
+      n--; NEXT_PRIME_VIADIFF(p,d);
+    } while (n) ;
+  }
+  return utoipos(p);
+}
+
+ulong
+uprime(long N)
+{
+  pari_sp av = avma;
+  GEN p;
+  if (N <= 0) pari_err_DOMAIN("prime", "n", "<=",gen_0, stoi(N));
+  p = prime_table_find_n(N);
+  if (lgefint(p) != 3) pari_err_OVERFLOW("uprime");
+  avma = av; return p[2];
 }
 GEN
-prime(long n) { return utoipos(uprime(n)); }
+prime(long N)
+{
+  pari_sp av = avma;
+  GEN p;
+  if (N <= 0) pari_err_DOMAIN("prime", "n", "<=",gen_0, stoi(N));
+  new_chunk(4); /*HACK*/
+  p = prime_table_find_n(N);
+  avma = av; return icopy(p);
+}
+
+/* random b-bit prime */
+GEN
+randomprime(GEN N)
+{
+  pari_sp av = avma, av2;
+  GEN a, b, d;
+  if (!N)
+    for(;;)
+    {
+      ulong p = random_bits(31);
+      if (uisprime(p)) return utoipos(p);
+    }
+  switch(typ(N))
+  {
+    case t_INT:
+      a = gen_2;
+      b = subiu(N,1); /* between 2 and N-1 */
+      d = subiu(N,2);
+      if (signe(d) <= 0)
+        pari_err_DOMAIN("randomprime","N", "<", gen_2, N);
+      break;
+    case t_VEC:
+      if (lg(N) != 3) pari_err_TYPE("randomprime",N);
+      a = gel(N,1);
+      b = gel(N,2);
+      if (gcmp(b, a) < 0)
+        pari_err_DOMAIN("randomprime","b-a", "<", gen_0, mkvec2(a,b));
+      if (typ(a) != t_INT)
+      {
+        a = gceil(a);
+        if (typ(a) != t_INT) pari_err_TYPE("randomprime",a);
+      }
+      if (typ(b) != t_INT)
+      {
+        b = gfloor(b);
+        if (typ(b) != t_INT) pari_err_TYPE("randomprime",b);
+      }
+      if (cmpis(a, 2) < 0)
+      {
+        a = gen_2;
+        d = subiu(b,1);
+      }
+      else
+        d = addiu(subii(b,a), 1);
+      if (signe(d) <= 0)
+        pari_err_DOMAIN("randomprime","floor(b) - max(ceil(a),2)", "<",
+                        gen_0, mkvec2(a,b));
+      break;
+    default:
+      pari_err_TYPE("randomprime", N);
+      return NULL; /*notreached*/
+  }
+  av2 = avma;
+  for (;;)
+  {
+    GEN p = addii(a, randomi(d));
+    if (BPSW_psp(p)) return gerepileuptoint(av, p);
+    avma = av2;
+  }
+}
+
+/* set *pp = nextprime(a) = p
+ *     *pd so that NEXT_PRIME_VIADIFF(d, p) = nextprime(p+1)
+ *     *pn so that p = the n-th prime
+ * error if nextprime(a) is out of primetable bounds */
+void
+prime_table_next_p(ulong a, byteptr *pd, ulong *pp, ulong *pn)
+{
+  byteptr d;
+  ulong p, n, maxp = maxprime();
+  long i = prime_table_closest_p(a);
+  p = prime_table[i].p;
+  if (p > a && p > maxp)
+  {
+    i--;
+    p = prime_table[i].p;
+  }
+  /* if beyond prime table, then p <= a */
+  n = prime_table[i].n;
+  d = diffptr + n;
+  if (p < a)
+  {
+    if (a > maxp) pari_err_MAXPRIME(a);
+    do { n++; NEXT_PRIME_VIADIFF(p,d); } while (p < a);
+  }
+  else if (p != a)
+  {
+    do { n--; PREC_PRIME_VIADIFF(p,d); } while (p > a) ;
+    if (p < a) { NEXT_PRIME_VIADIFF(p,d); n++; }
+  }
+  *pn = n;
+  *pp = p;
+  *pd = d;
+}
 
 ulong
-uprimepi(ulong n)
+uprimepi(ulong a)
 {
-  byteptr p = diffptr+1;
-  ulong prime = 2, res = 0;
-  maxprime_check(n);
-  while (prime < n) { res++; NEXT_PRIME_VIADIFF(prime,p); }
-  return prime == n? res+1: res;
+  ulong p, n, maxp = maxprime();
+  if (a <= maxp)
+  {
+    byteptr d;
+    prime_table_next_p(a, &d, &p, &n);
+    return p == a? n: n-1;
+  }
+  else
+  {
+    long i = prime_table_closest_p(a);
+    forprime_t S;
+    p = prime_table[i].p;
+    if (p > a)
+    {
+      i--;
+      p = prime_table[i].p;
+    }
+    /* p = largest prime in table <= a */
+    n = prime_table[i].n;
+    (void)u_forprime_init(&S, p+1, a);
+    for (; p; n++) p = u_forprime_next(&S);
+    return n-1;
+  }
 }
 
 GEN
 primepi(GEN x)
 {
   pari_sp av = avma;
-  GEN N = typ(x) == t_INT? x: gfloor(x);
-  if (typ(N) != t_INT) pari_err(typeer, "primepi");
+  GEN pp, nn, N = typ(x) == t_INT? x: gfloor(x);
+  forprime_t S;
+  ulong n, p;
+  long i, l;
+  if (typ(N) != t_INT) pari_err_TYPE("primepi",N);
   if (signe(N) <= 0) return gen_0;
-  avma = av; return utoi(uprimepi(itou(N)));
+  avma = av; l = lgefint(N);
+  if (l == 3) return utoi(uprimepi(N[2]));
+  i = prime_table_len-1;
+  p = prime_table[i].p;
+  n = prime_table[i].n;
+  (void)forprime_init(&S, utoipos(p+1), N);
+  nn = setloop(utoipos(n));
+  pp = gen_0;
+  for (; pp; incloop(nn)) pp = forprime_next(&S);
+  return gerepileuptoint(av, subiu(nn,1));
+}
+
+/* pi(x) < x/log x * (1 + 1/log x + 2.51/log^2 x)), x>=355991 [ Dusart ]
+ * pi(x) < x/(log x - 1.1), x >= 60184 [ Dusart ]
+ * ? \p9
+ * ? M = 0; for(x = 4, 60184, M = max(M, log(x) - x/primepi(x))); M
+ * %1 = 1.11196252 */
+double
+primepi_upper_bound(double x)
+{
+  if (x >= 355991)
+  {
+    double L = 1/log(x);
+    return x * L * (1 + L + 2.51*L*L);
+  }
+  if (x >= 60184) return x / (log(x) - 1.1);
+  if (x < 5) return 2; /* don't bother */
+  return x / (log(x) - 1.111963);
+}
+/* pi(x) > x/log x (1 + 1/log x), x >= 599 [ Dusart ]
+ * pi(x) > x / (log x + 2), x >= 55 [ Rosser ] */
+double
+primepi_lower_bound(double x)
+{
+  if (x >= 599)
+  {
+    double L = 1/log(x);
+    return x * L * (1 + L);
+  }
+  if (x < 55) return 0; /* don't bother */
+  return x / (log(x) + 2.);
+}
+GEN
+gprimepi_upper_bound(GEN x)
+{
+  pari_sp av = avma;
+  double L;
+  if (typ(x) != t_INT) x = gfloor(x);
+  if (expi(x) <= 1022)
+  {
+    avma = av;
+    return dbltor(primepi_upper_bound(gtodouble(x)));
+  }
+  x = itor(x, LOWDEFAULTPREC);
+  L = 1 / rtodbl(logr_abs(x));
+  x = mulrr(x, dbltor(L * (1 + L + 2.51*L*L)));
+  return gerepileuptoleaf(av, x);
+}
+GEN
+gprimepi_lower_bound(GEN x)
+{
+  pari_sp av = avma;
+  double L;
+  if (typ(x) != t_INT) x = gfloor(x);
+  if (cmpiu(x, 55) <= 0) return gen_0;
+  if (expi(x) <= 1022)
+  {
+    avma = av;
+    return dbltor(primepi_lower_bound(gtodouble(x)));
+  }
+  x = itor(x, LOWDEFAULTPREC);
+  L = 1 / rtodbl(logr_abs(x));
+  x = mulrr(x, dbltor(L * (1 + L)));
+  return gerepileuptoleaf(av, x);
 }
 
 GEN
-primes(long m)
+primes(long n)
 {
-  byteptr p = diffptr;
-  ulong prime = 0;
-  long n = (m < 0)? 0: m;
-  GEN y, z;
-
-  z = y = cgetg(n+1,t_VEC);
-  while (n--)
-  {
-    NEXT_PRIME_VIADIFF(prime,p);
-    if (!*p) /* use something close to Dusart's bound */
-      maxprime_check((ulong)(m*( log((double)m*log((double)m))-0.948 )));
-    gel(++z, 0) = utoipos(prime);
-  }
+  forprime_t S;
+  long i;
+  GEN y;
+  if (n <= 0) return cgetg(1, t_VEC);
+  y = cgetg(n+1, t_VEC);
+  (void)new_chunk(3*n); /*HACK*/
+  u_forprime_init(&S, 2, ULONG_MAX);
+  avma = (pari_sp)y;
+  for (i = 1; i <= n; i++) gel(y, i) = utoipos( u_forprime_next(&S) );
   return y;
 }
 GEN
-primes_zv(long m)
+primes_zv(long n)
 {
-  byteptr p = diffptr;
-  ulong prime = 0;
-  long n = (m < 0)? 0: m;
-  GEN y, z;
-
-  z = y = cgetg(n+1,t_VECSMALL);
-  while (n--)
+  forprime_t S;
+  long i;
+  GEN y;
+  if (n <= 0) return cgetg(1, t_VECSMALL);
+  y = cgetg(n+1,t_VECSMALL);
+  u_forprime_init(&S, 2, ULONG_MAX);
+  for (i = 1; i <= n; i++) y[i] =  u_forprime_next(&S);
+  avma = (pari_sp)y; return y;
+}
+GEN
+primes0(GEN N)
+{
+  switch(typ(N))
   {
-    NEXT_PRIME_VIADIFF(prime,p);
-    if (!*p) /* use something close to Dusart's bound */
-      maxprime_check((ulong)(m*( log((double)m*log((double)m))-0.948 )));
-    *++z = prime;
+    case t_INT: return primes(itos(N));
+    case t_VEC:
+      if (lg(N) == 3) return primes_interval(gel(N,1),gel(N,2));
   }
+  pari_err_TYPE("primes", N);
+  return NULL;
+}
+
+GEN
+primes_interval(GEN a, GEN b)
+{
+  pari_sp av = avma;
+  forprime_t S;
+  long i, n;
+  GEN y, d, p;
+  if (typ(a) != t_INT) a = gceil(a);
+  if (typ(b) != t_INT) b = gfloor(b);
+  d = subii(b, a);
+  if (signe(d) < 0 || signe(b) <= 0) { avma = av; return cgetg(1, t_VEC); }
+  if (lgefint(b) == 3)
+  {
+    avma = av;
+    y = primes_interval_zv(itou(a), itou(b));
+    n = lg(y); settyp(y, t_VEC);
+    for (i = 1; i < n; i++) gel(y,i) = utoipos(y[i]);
+    return y;
+  }
+  /* at most d+1 primes in [a,b]. If d large, try better bound to lower
+   * memory use */
+  if (cmpiu(d,100000) > 0)
+  {
+    GEN D = gsub(gprimepi_upper_bound(b), gprimepi_lower_bound(a));
+    D = ceil_safe(D);
+    if (cmpii(D, d) < 0) d = D;
+  }
+  n = itos(d)+1;
+  forprime_init(&S, a, b);
+  y = cgetg(n+1, t_VEC); i = 1;
+  while ((p = forprime_next(&S))) gel(y, i++) = icopy(p);
+  setlg(y, i); return gerepileupto(av, y);
+}
+
+/* a <= b, at most d primes in [a,b]. Return them */
+static GEN
+primes_interval_i(ulong a, ulong b, ulong d)
+{
+  ulong p, i = 1, n = d + 1;
+  forprime_t S;
+  GEN y = cgetg(n+1, t_VECSMALL);
+  pari_sp av = avma;
+  u_forprime_init(&S, a, b);
+  while ((p = u_forprime_next(&S))) y[i++] = p;
+  avma = av; setlg(y, i); stackdummy((pari_sp)(y + i), (pari_sp)(y + n+1));
   return y;
 }
+GEN
+primes_interval_zv(ulong a, ulong b)
+{
+  ulong d;
+  if (!a) return primes_upto_zv(b);
+  if (b < a) return cgetg(1, t_VECSMALL);
+  d = b - a;
+  if (d > 100000UL) d = primepi_upper_bound(b) - primepi_lower_bound(a);
+  return primes_interval_i(a, b, d);
+}
+GEN
+primes_upto_zv(ulong b)
+{
+  ulong d;
+  if (b < 2) return cgetg(1, t_VECSMALL);
+  d = (b > 100000UL)? primepi_upper_bound(b): b;
+  return primes_interval_i(2, b, d);
+}
 
 /***********************************************************************/
 /**                                                                   **/
@@ -837,9 +1261,11 @@ static void
 rmprime(GEN T, GEN p)
 {
   long i;
-  if (typ(p) != t_INT) pari_err(typeer,"rmprime");
+  if (typ(p) != t_INT) pari_err_TYPE("removeprimes",p);
   i = ZV_search(T, p);
-  if (!i) pari_err(talker,"prime %Ps is not in primetable", p);
+  if (!i)
+    pari_err_DOMAIN("removeprime","prime","not in",
+                    strtoGENstr("primetable"), p);
   gunclone(gel(T,i)); gel(T,i) = NULL;
   cleanprimetab(T);
 }
@@ -883,8 +1309,7 @@ addp(GEN *T, GEN p)
   RgV_check_ZV(p, "addprimes");
   v = gen_indexsort_uniq(p, (void*)&cmpii, &cmp_nodata);
   p = vecpermute(p, v);
-  if (cmpii(gel(p,1), gen_1) <= 0)
-    pari_err(talker,"entries must be > 1 in addprimes");
+  if (cmpiu(gel(p,1), 2) < 0) pari_err_DOMAIN("addprimes", "p", "<", gen_2,p);
   p = addp_union(*T, p);
   l = lg(p);
   if (l != lg(*T))
