@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -130,7 +128,6 @@ struct data_qt
 static void SetForeground(void *data, long col)
 {
    struct data_qt *d = (struct data_qt *) data;
-   if (col >= d->numcolors) col = d->numcolors-1;
    d->p->setPen(d->color[col]);
 }
 
@@ -192,9 +189,10 @@ void Plotter::draw(QPainter *p){
   plotQt.ml=&DrawLines;
   plotQt.st=&DrawString;
   plotQt.pl=&pari_plot;
+  plotQt.data = (void *)&d;
   double xs = double(this->width()) / pari_plot.width,
          ys = double(this->height()) / pari_plot.height;
-  gen_rectdraw0(&plotQt, (void *)&d, this->w, this->x, this->y,this->lw,xs,ys);
+  gen_rectdraw0(&plotQt, this->w, this->x, this->y,this->lw,xs,ys);
 }
 
 void Plotter::save( const QString& s, const QString& f){
@@ -581,7 +579,7 @@ rectdraw0(long *w, long *x, long *y, long lw)
     if (pari_daemon()) return;  // parent process returns
 
     pari_close();
-    PARI_get_plot(1);
+    PARI_get_plot();
 
     // launch Qt window
     int argc = 1; char *argv[] = { "gp", "-qws"}; // set argc = 2 for cross
@@ -608,10 +606,9 @@ rectdraw0(long *w, long *x, long *y, long lw)
 }
 
 void
-PARI_get_plot(long f)
+PARI_get_plot()
 /* This function initialises the structure rect.h: pari_plot */
 {
-    (void)f;
     if (pari_plot.init) return;      // pari_plot is already set
 #ifdef __QPE__
     pari_plot.width   = 240;         // width and
