@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2009  The PARI group.
+/* Copyright (C) 2009  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -92,19 +90,19 @@ void rectdraw0(long *w, long *x, long *y, long lw)
   plotWin32.ml=&DrawLines;
   plotWin32.st=&DrawString;
   plotWin32.pl=&pari_plot;
+  plotWin32.data=(void*)hEmf;
 
-  gen_rectdraw0(&plotWin32, (void*)hEmf, w, x, y, lw, 1, 1);
+  gen_rectdraw0(&plotWin32, w, x, y, lw, 1, 1);
   DeleteEnhMetaFile(CloseEnhMetaFile(hEmf));
 
   ShellExecute(NULL,NULL,fname,NULL,NULL,SW_SHOWDEFAULT);
 }
 
 void
-PARI_get_plot(long f)
+PARI_get_plot()
 {
   HDC hdc;
   TEXTMETRIC tm;
-  (void) f;
   if (pari_plot.init) return;      /* pari_plot is already set */
 
   pari_plot.init    = 1;
