@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000  The PARI group.
+/* Copyright (C) 2000  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -26,8 +24,5 @@ rectdraw0(long *w, long *x, long *y, long lw)
 }
 
 void
-PARI_get_plot(long f)
-{
-  (void)f;
-  pari_err(talker,"high resolution graphics disabled");
-}
+PARI_get_plot()
+{ pari_err(e_MISC,"high resolution graphics disabled"); }
