@@ -1,7 +1,5 @@
 #line 2 "../src/kernel/none/ratlift.c"
-/* $Id$
-
-Copyright (C) 2003  The PARI group.
+/* Copyright (C) 2003  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -15,7 +13,7 @@ with the package; see the file 'COPYING'. If not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 
 /*==========================================================
- * ratlift(GEN x, GEN m, GEN *a, GEN *b, GEN amax, GEN bmax)
+ * Fp_ratlift(GEN x, GEN m, GEN *a, GEN *b, GEN amax, GEN bmax)
  *==========================================================
  * Reconstruct rational number from its residue x mod m
  *    Given t_INT x, m, amax>=0, bmax>0 such that
@@ -44,30 +42,9 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
       PAGES = {287--297},
  }
  * Preprint available from:
- * ftp://ftp.risc.uni-linz.ac.at/pub/techreports/1994/94-64.ps.gz
- */
+ * ftp://ftp.risc.uni-linz.ac.at/pub/techreports/1994/94-64.ps.gz */
 
 /* #define DEBUG_RATLIFT */
-
-int
-ratlift(GEN x, GEN m, GEN amax, GEN bmax, GEN *a, GEN *b)
-{
-  pari_sp av = avma;
-  if ((typ(x) | typ(m) | typ(amax) | typ(bmax)) != t_INT) pari_err(arither1);
-  if (signe(bmax) <= 0)
-    pari_err(talker, "ratlift: bmax must be > 0, found\n\tbmax=%Ps\n", bmax);
-  if (signe(amax) < 0)
-    pari_err(talker, "ratilft: amax must be >= 0, found\n\tamax=%Ps\n", amax);
-  /* check 2*amax*bmax < m */
-  if (cmpii(shifti(mulii(amax, bmax), 1), m) >= 0)
-    pari_err(talker, "ratlift: must have 2*amax*bmax < m, found\n\tamax=%Ps\n\tbmax=%Ps\n\tm=%Ps\n", amax,bmax,m);
-  /* we _could_ silently replace x with modii(x,m) instead of the following,
-   * but let's leave this up to the caller */
-  if (signe(x) < 0 || cmpii(x,m) >= 0)
-    pari_err(talker, "ratlift: must have 0 <= x < m, found\n\tx=%Ps\n\tm=%Ps\n", x,m);
-  avma = av; return Fp_ratlift(x, m, amax, bmax, a, b);
-}
-
 static ulong
 get_vmax(GEN r, long lb, long lbb)
 {
@@ -307,6 +284,6 @@ Fp_ratlift(GEN x, GEN m, GEN amax, GEN bmax, GEN *a, GEN *b)
   } /* while */
 
   /* We have run into d1 == 0 before returning. This cannot happen */
-  pari_err(bugparier, "ratlift failed to catch d1 == 0\n");
+  pari_err_BUG("ratlift failed to catch d1 == 0");
   return 0; /* NOTREACHED */
 }
