@@ -477,8 +477,8 @@ pop_lex(long n)
   s_trace.n--;
 }
 
-static pari_stack s_relocs;
-static entree **relocs;
+static THREAD pari_stack s_relocs;
+static THREAD entree **relocs;
 
 void
 pari_init_evaluator(void)
@@ -950,18 +950,21 @@ closure_eval(GEN C)
       break;
     case OCgerepile:
     {
-      pari_sp av, av2;
+      pari_sp av;
       GEN x;
       sp--;
       av = st[sp-1];
       x = gel(st,sp);
-      av2 = (pari_sp)(x + lg(x));
-      if (av - av2 > 1000000)
+      if (isonstack(x))
       {
-        if (DEBUGMEM>=2)
-          pari_warn(warnmem,"eval: recovering %ld bytes", av - av2);
-        x = gerepileupto(av, x);
-      }
+        pari_sp av2 = (pari_sp)(x + lg(x));
+        if ((long) (av - av2) > 1000000L)
+        {
+          if (DEBUGMEM>=2)
+            pari_warn(warnmem,"eval: recovering %ld bytes", av - av2);
+          x = gerepileupto(av, x);
+        }
+      } else avma = av;
       gel(st,sp-1) = x;
       break;
     }
@@ -1396,6 +1399,14 @@ evalstate_restore(struct pari_evalstate *state)
   compilestate_restore(&state->comp);
 }
 
+GEN
+evalstate_restore_err(struct pari_evalstate *state)
+{
+  GENbin* err = copy_bin(pari_err_last());
+  evalstate_restore(state);
+  return bin_copy(err);
+}
+
 void
 evalstate_reset(void)
 {
