@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2000-2003  The PARI group.
+/* Copyright (C) 2000-2003  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -21,53 +19,63 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA. */
 /**                          MEMBER FUNCTIONS                      **/
 /**                                                                **/
 /********************************************************************/
-INLINE int is_ell5(GEN x) {
-  long lx = lg(x);
-  return (typ(x) == t_VEC && (lx == 6 || lx == 14 || lx == 20));
-}
-INLINE int is_smallell(GEN x) {
-  long lx = lg(x);
-  return (typ(x) == t_VEC && (lx == 14 || lx == 20));
+INLINE int
+is_ell5(GEN x) {
+  long lx;
+  if (typ(x) != t_VEC) return 0;
+  lx = lg(x);
+  return lx == 17 || (lx == 6 && !is_vec_t(typ(gel(x,2))));
 }
 INLINE int is_ell(GEN x) {
   long lx = lg(x);
-  return (typ(x) == t_VEC && lx == 20);
+  return (typ(x) == t_VEC && lx == 17);
 }
 
 static void
-member_err(const char *s) { pari_err(typeer,s); }
+member_err(const char *s, GEN y) { pari_err_TYPE(s,y); }
 
 GEN
 member_e(GEN x)
 {
-  x = get_prid(x);
-  if (!x) member_err("e");
-  return gel(x,3);
+  GEN y = get_prid(x);
+  if (!y) member_err("e",x);
+  return gel(y,3);
 }
 
 GEN
 member_f(GEN x)
 {
-  x = get_prid(x);
-  if (!x) member_err("f");
-  return gel(x,4);
+  GEN y = get_prid(x);
+  if (!y)
+  {
+    if (typ(x) == t_FFELT) return utoipos(FF_f(x));
+    member_err("f",x);
+  }
+  return gel(y,4);
 }
 
 GEN
 member_p(GEN x)
 {
   long t; (void)get_nf(x,&t);
-  if (t == typ_GAL) return gal_get_p(x);
-  switch(typ(x)) {
-    case t_VEC:
-      x = get_prid(x); if (!x) member_err("p");
-      return pr_get_p(x);
-    case t_PADIC:
-      return gel(x,2);
-    case t_FFELT:
-      return FF_p_i(x);
+  switch(t)
+  {
+    case typ_GAL: return gal_get_p(x);
+    case typ_ELL: switch(ell_get_type(x))
+    {
+      case t_ELL_Fp:
+      case t_ELL_Fq: return ellff_get_p(x);
+      case t_ELL_Qp: return ellQp_get_p(x);
+      default: member_err("p",x);
+    }
+    case typ_MODPR: x = get_prid(x);
+    case typ_PRID: return pr_get_p(x);
   }
-  member_err("p");
+  switch(typ(x)) {
+    case t_PADIC: return gel(x,2);
+    case t_FFELT: return FF_p_i(x);
+  }
+  member_err("p",x);
   return NULL;
 }
 
@@ -79,16 +87,16 @@ member_bid(GEN x)
     case typ_BNR: return bnr_get_bid(x);
     case typ_BID: return x;
   }
-  member_err("bid");
+  member_err("bid",x);
   return NULL;
 }
 
 GEN
 member_bnf(GEN x)
 {
-  long t; x = get_bnf(x,&t);
-  if (!x) member_err("bnf");
-  return x;
+  long t; GEN y = get_bnf(x,&t);
+  if (!y) member_err("bnf",x);
+  return y;
 }
 
 GEN
@@ -97,7 +105,7 @@ member_nf(GEN x)
   long t; GEN y = get_nf(x,&t);
   if (!y) {
     if (t == typ_RNF) return gel(x,10);
-    member_err("nf");
+    member_err("nf",x);
   }
   return y;
 }
@@ -114,11 +122,11 @@ member_zk(GEN x)
       case typ_Q:
         y = cgetg(3,t_VEC);
         gel(y,1) = gen_1;
-        gel(y,2) = pol_x(varn(x[1])); return y;
+        gel(y,2) = pol_x(varn(gel(x,1))); return y;
       case typ_RNF:
         return gel(x,7);
     }
-    member_err("zk");
+    member_err("zk",x);
   }
   return nf_get_zk(y);
 }
@@ -133,8 +141,9 @@ member_disc(GEN x) /* discriminant */
     {
       case typ_Q  : return quad_disc(x);
       case typ_ELL: return ell_get_disc(x);
+      case typ_RNF: return rnf_get_disc(x);
     }
-    member_err("disc");
+    member_err("disc",x);
   }
   return nf_get_disc(y);
 }
@@ -150,15 +159,23 @@ member_pol(GEN x) /* polynomial */
       case typ_POL: return x;
       case typ_Q  : return gel(x,1);
       case typ_GAL: return gal_get_pol(x);
-      case typ_RNF: return gmael(x,11,1);
+      case typ_RNF: return rnf_get_pol(x);
     }
     if (typ(x)==t_POLMOD) return gel(x,2);
     if (typ(x)==t_FFELT) return FF_to_FpXQ(x);
-    member_err("pol");
+    member_err("pol",x);
   }
   return nf_get_pol(y);
 }
 
+GEN
+member_polabs(GEN x)
+{
+  long t; (void)get_nf(x,&t);
+  if (t != typ_RNF) member_err("pol",x);
+  return rnf_get_polabs(x);
+}
+
 GEN
 member_mod(GEN x) /* modulus */
 {
@@ -173,7 +190,7 @@ member_mod(GEN x) /* modulus */
     case t_INTMOD: case t_POLMOD: case t_QUAD: break;
     case t_PADIC: return gel(x,3);
     case t_FFELT: return FF_mod(x);
-    default: member_err("mod");
+    default: member_err("mod",x);
   }
   return gel(x,1);
 }
@@ -182,7 +199,7 @@ GEN
 member_sign(GEN x) /* signature */
 {
   long t; GEN y = get_nf(x,&t);
-  if (!y) member_err("sign");
+  if (!y) member_err("sign",x);
   return gel(y,2);
 }
 GEN
@@ -194,7 +211,11 @@ GEN
 member_index(GEN x)
 {
   long t; GEN y = get_nf(x,&t);
-  if (!y) member_err("index");
+  if (!y)
+  {
+    if (t == typ_RNF) return rnf_get_index(x);
+    member_err("index",x);
+  }
   return nf_get_index(y);
 }
 
@@ -205,31 +226,31 @@ nfmats(GEN x)
   GEN y;
   if (!x) return NULL;
   y = gel(x,5);
-  if (typ(y) == t_VEC && lg(y) != 8) return NULL;
+  if (typ(y) == t_VEC && lg(y) < 8) return NULL;
   return y;
 }
 
 GEN
 member_t2(GEN x) /* T2 matrix */
 {
-  long t; x = nfmats(get_nf(x,&t));
-  if (!x) member_err("t2");
-  return gram_matrix(gel(x,2));
+  long t; GEN y = nfmats(get_nf(x,&t));
+  if (!y) member_err("t2",x);
+  return gram_matrix(gel(y,2));
 }
 
 GEN
 member_diff(GEN x) /* different */
 {
-  long t; x = nfmats(get_nf(x,&t));
-  if (!x) member_err("diff");
-  return gel(x,5);
+  long t; GEN y = nfmats(get_nf(x,&t));
+  if (!y) member_err("diff",x);
+  return gel(y,5);
 }
 
 GEN
 member_codiff(GEN x) /* codifferent */
 {
   long t; GEN T, D, DinvT, nf = get_nf(x,&t), y = nfmats(nf);
-  if (!y) member_err("codiff");
+  if (!y) member_err("codiff",x);
   T = gel(y,4);
   D = absi(nf_get_disc(nf));
   DinvT = ZM_inv(T,D);
@@ -242,9 +263,15 @@ member_roots(GEN x) /* roots */
   long t; GEN y = get_nf(x,&t);
   if (!y)
   {
-    if (t == typ_ELL && is_ell(x)) return ell_get_roots(x);
     if (t == typ_GAL) return gal_get_roots(x);
-    member_err("roots");
+    if (t == typ_ELL)
+      switch(ell_get_type(x))
+      {
+        case t_ELL_Qp: return mkcol( ellQp_root(x, ellQp_get_prec(x)) );
+        case t_ELL_Q:
+        case t_ELL_Rg: return ellR_roots(x, ellR_get_prec(x));
+      }
+    member_err("roots",x);
   }
   return nf_get_roots(y);
 }
@@ -254,14 +281,14 @@ static GEN
 check_RES(GEN x, const char *s)
 {
   GEN y = gel(x,8);
-  if (typ(y) != t_VEC || lg(y) < 4) member_err(s);
+  if (typ(y) != t_VEC || lg(y) < 4) member_err(s,x);
   return y;
 }
 
-GEN
-member_clgp(GEN x) /* class group (3-component row vector) */
+/* y = get_bnf(x, &t) */
+static GEN
+_member_clgp(GEN x, GEN y, long t) /* class group (3-component row vector) */
 {
-  long t; GEN y = get_bnf(x,&t);
   if (!y)
   {
     switch(t)
@@ -275,12 +302,17 @@ member_clgp(GEN x) /* class group (3-component row vector) */
         case 3: /* no gen */
         case 4: return x;
       }
-    member_err("clgp");
+    member_err("clgp",x);
   }
   if (t==typ_BNR) return gel(x,5);
-  y = check_RES(y, "clgp");
-  return gel(y,1);
+  y = check_RES(y, "clgp"); return gel(y,1);
 }
+static GEN
+_check_clgp(GEN x, GEN y, long t)
+{ GEN c = _member_clgp(x,y,t); checkabgrp(c); return c; }
+GEN
+member_clgp(GEN x)
+{ long t; GEN y = get_bnf(x,&t); return _check_clgp(x,y,t); }
 
 GEN
 member_reg(GEN x) /* regulator */
@@ -289,9 +321,9 @@ member_reg(GEN x) /* regulator */
   if (!y)
   {
     if (t == typ_QUA) return gel(x,4);
-    member_err("reg");
+    member_err("reg",x);
   }
-  if (t == typ_BNR) pari_err(impl,"ray regulator");
+  if (t == typ_BNR) pari_err_IMPL("ray regulator");
   y = check_RES(y, "reg");
   return gel(y,2);
 }
@@ -308,9 +340,9 @@ member_fu(GEN x) /* fundamental units */
         x = quad_disc(x);
         return (signe(x)<0)? cgetg(1,t_VEC): quadunit(x);
     }
-    member_err("fu");
+    member_err("fu",x);
   }
-  if (t == typ_BNR) pari_err(impl,"ray units");
+  if (t == typ_BNR) pari_err_IMPL("ray units");
   return matbasistoalg(y, bnf_get_fu(y));
 }
 
@@ -323,7 +355,7 @@ member_tu(GEN x)
   if (!bnf)
   {
     GEN y;
-    if (t != typ_Q) member_err("tu");
+    if (t != typ_Q) member_err("tu",x);
     y = quad_disc(x);
     if (signe(y) > 0 || cmpiu(y,4) > 0) return mkvec2(gen_m1, gen_2);
 
@@ -332,7 +364,7 @@ member_tu(GEN x)
   }
   else
   {
-    if (t == typ_BNR) pari_err(impl,"ray torsion units");
+    if (t == typ_BNR) pari_err_IMPL("ray torsion units");
     gel(res,1) = utoipos( bnf_get_tuN(bnf) );
     gel(res,2) = basistoalg(bnf, bnf_get_tuU(bnf));
   }
@@ -365,26 +397,36 @@ member_zkst(GEN x)
       if (typ(bid) == t_VEC && lg(bid) > 2) return gel(bid,2);
     }
   }
-  member_err("zkst");
+  member_err("zkst",x);
   return NULL; /* not reached */
 }
 
 GEN
-member_no(GEN clg) /* number of elements of a group (of type clgp) */
+member_no(GEN x) /* number of elements of a group (of type clgp) */
 {
   pari_sp av = avma;
-  clg = member_clgp(clg);
-  if (typ(clg)!=t_VEC || (lg(clg)!=3 && lg(clg)!=4)) member_err("no");
-  avma = av; return gel(clg,1);
+  long t; GEN y = get_bnf(x,&t);
+  if (t == typ_ELL) switch(ell_get_type(x))
+  {
+    case t_ELL_Fp:
+    case t_ELL_Fq: return ellcard(x, NULL);
+  }
+  x = _check_clgp(x,y,t);
+  avma = av; return gel(x,1);
 }
 
 GEN
-member_cyc(GEN clg) /* cyclic decomposition (SNF) of a group (of type clgp) */
+member_cyc(GEN x) /* cyclic decomposition (SNF) of a group (of type clgp) */
 {
   pari_sp av = avma;
-  clg = member_clgp(clg);
-  if (typ(clg)!=t_VEC || (lg(clg)!=3 && lg(clg)!=4)) member_err("cyc");
-  avma = av; return gel(clg,2);
+  long t; GEN y = get_bnf(x,&t);
+  if (t == typ_ELL) switch(ell_get_type(x))
+  {
+    case t_ELL_Fp:
+    case t_ELL_Fq: return ellgroup(x, NULL);
+  }
+  x = _check_clgp(x,y,t);
+  avma = av; return gel(x,2);
 }
 
 /* SNF generators of a group (of type clgp), or generators of a prime
@@ -393,24 +435,26 @@ GEN
 member_gen(GEN x)
 {
   pari_sp av;
-  long t;
-  GEN y = get_prid(x);
-  if (y) return mkvec2(gel(y,1), gel(y,2));
-  (void)get_nf(x,&t);
-  if (t == typ_GAL) return gal_get_gen(x);
+  long t; GEN y = get_bnf(x,&t);
+  switch(t)
+  {
+    case typ_MODPR: x = get_prid(x);
+    case typ_PRID: return mkvec2(gel(x,1), gel(x,2));
+    case typ_GAL: return gal_get_gen(x);
+    case typ_ELL: return ellgenerators(x);
+  }
   av = avma;
-  x = member_clgp(x);
-  if (typ(x)!=t_VEC || lg(x)!=4) member_err("gen");
-  avma = av;
-  if (typ(x[1]) == t_COL) return gel(x,2); /* from bnfisprincipal */
-  return gel(x,3);
+  x = _check_clgp(x,y,t);
+  if (lg(x)!=4) member_err("gen",x);
+  avma = av; return gel(x,3);
 }
 GEN
 member_group(GEN x)
 {
   long t; (void)get_nf(x,&t);
   if (t == typ_GAL) return gal_get_group(x);
-  member_err("group");
+  if (t == typ_ELL) return ellgroup0(x, NULL, 1);
+  member_err("group",x);
   return NULL; /* not reached */
 }
 GEN
@@ -418,132 +462,135 @@ member_orders(GEN x)
 {
   long t; (void)get_nf(x,&t);
   if (t == typ_GAL) return gal_get_orders(x);
-  member_err("orders");
+  member_err("orders",x);
   return NULL; /* not reached */
 }
 
 GEN
 member_a1(GEN x)
 {
-  if (!is_ell5(x)) member_err("a1");
+  if (!is_ell5(x)) member_err("a1",x);
   return ell_get_a1(x);
 }
 
 GEN
 member_a2(GEN x)
 {
-  if (!is_ell5(x)) member_err("a2");
+  if (!is_ell5(x)) member_err("a2",x);
   return ell_get_a2(x);
 }
 
 GEN
 member_a3(GEN x)
 {
-  if (!is_ell5(x)) member_err("a3");
+  if (!is_ell5(x)) member_err("a3",x);
   return ell_get_a3(x);
 }
 
 GEN
 member_a4(GEN x)
 {
-  if (!is_ell5(x)) member_err("a4");
+  if (!is_ell5(x)) member_err("a4",x);
   return ell_get_a4(x);
 }
 
 GEN
 member_a6(GEN x)
 {
-  if (!is_ell5(x)) member_err("a6");
+  if (!is_ell5(x)) member_err("a6",x);
   return ell_get_a6(x);
 }
 
 GEN
 member_b2(GEN x)
 {
-  if (!is_smallell(x)) member_err("b2");
+  if (!is_ell(x)) member_err("b2",x);
   return ell_get_b2(x);
 }
 
 GEN
 member_b4(GEN x)
 {
-  if (!is_smallell(x)) member_err("b4");
+  if (!is_ell(x)) member_err("b4",x);
   return ell_get_b4(x);
 }
 
 GEN
 member_b6(GEN x)
 {
-  if (!is_smallell(x)) member_err("b6");
+  if (!is_ell(x)) member_err("b6",x);
   return ell_get_b6(x);
 }
 
 GEN
 member_b8(GEN x)
 {
-  if (!is_smallell(x)) member_err("b8");
+  if (!is_ell(x)) member_err("b8",x);
   return ell_get_b8(x);
 }
 
 GEN
 member_c4(GEN x)
 {
-  if (!is_smallell(x)) member_err("c4");
+  if (!is_ell(x)) member_err("c4",x);
   return ell_get_c4(x);
 }
 
 GEN
 member_c6(GEN x)
 {
-  if (!is_smallell(x)) member_err("c6");
+  if (!is_ell(x)) member_err("c6",x);
   return ell_get_c6(x);
 }
 
 GEN
 member_j(GEN x)
 {
-  if (!is_smallell(x)) member_err("j");
+  if (!is_ell(x)) member_err("j",x);
   return ell_get_j(x);
 }
 
+static int
+ell_is_complex(GEN x)
+{ long t = ell_get_type(x); return t == t_ELL_Q || t == t_ELL_Rg; }
+
 GEN
 member_omega(GEN x)
 {
-  if (!is_ell(x)) member_err("omega");
-  if (!ell_is_real(x)) pari_err(talker,"curve not defined over R");
-  return mkvec2(gel(x,15), gel(x,16));
+  if (!is_ell(x)) member_err("omega",x);
+  if (!ell_is_complex(x)) pari_err_TYPE("omega [not defined over C]",x);
+  return ellR_omega(x, ellR_get_prec(x));
 }
 
 GEN
 member_eta(GEN x)
 {
-  if (!is_ell(x)) member_err("eta");
-  if (!ell_is_real(x)) pari_err(talker,"curve not defined over R");
-  return mkvec2(gel(x,17), gel(x,18));
+  if (!is_ell(x)) member_err("eta",x);
+  if (!ell_is_complex(x)) pari_err_TYPE("eta [not defined over C]",x);
+  return ellR_eta(x, ellR_get_prec(x));
 }
 
 GEN
 member_area(GEN x)
 {
-  if (!is_ell(x)) member_err("area");
-  if (!ell_is_real(x)) pari_err(talker,"curve not defined over R");
-  return gel(x,19);
+  GEN w, w1, w2, a,b,c,d;
+  long prec;
+  if (!is_ell(x)) member_err("area",x);
+  if (!ell_is_complex(x)) pari_err_TYPE("area [not defined over C]",x);
+  prec = ellR_get_prec(x);
+  w = ellR_omega(x, prec);
+  w1 = gel(w,1); a = real_i(w1); b = imag_i(w1);
+  w2 = gel(w,2); c = real_i(w2); d = imag_i(w2);
+  return gabs(gsub(gmul(a,d),gmul(b,c)), prec);
 }
 
 GEN
 member_tate(GEN x)
 {
-  if (!is_ell(x)) member_err("tate");
-  if (!ell_is_padic(x))
-    pari_err(talker,"curve not defined over a p-adic field");
-  return mkvec3(gel(x,15), gel(x,16), gel(x,17));
-}
-
-GEN
-member_w(GEN x)
-{
-  if (!is_ell(x)) member_err("w");
-  if (!ell_is_padic(x))
-    pari_err(talker,"curve not defined over a p-adic field");
-  return gel(x,18);
+  long prec;
+  if (!is_ell(x)) member_err("tate",x);
+  if (ell_get_type(x) != t_ELL_Qp)
+    pari_err_TYPE("tate [not defined over Qp]",x);
+  prec = ellQp_get_prec(x);
+  return ellQp_Tate_uniformization(x, prec);
 }
