@@ -1,6 +1,4 @@
-/* $Id$
-
-Copyright (C) 2009  The PARI group.
+/* Copyright (C) 2009  The PARI group.
 
 This file is part of the PARI/GP package.
 
@@ -316,7 +314,7 @@ GetSharp(FAD_t *fdata, GEN pp, GEN ppp, GEN pol, GEN beta, long *pl)
     if (i <= v) break;
     p1 = shallowcopy(p2);
   }
-  if (!v) pari_err(talker, "No division in GetSharp");
+  if (!v) pari_err_BUG("GetSharp [no division]");
 
   for (l = v; l >= 0; l--)
   {
@@ -409,7 +407,7 @@ RootCongruents(KRASNER_t *data, FAD_t *fdata, GEN pol, GEN alpha, GEN pp, GEN pp
     if (l <= 1) return l;
 #ifdef CHECK_EXTENSIONS
     if (l != DegreeMod(fdata, pp, ppp, pol))
-      pari_err(talker, "Degrees do not match in RCA");
+      pari_err_BUG("RootCongruents [degree mismatch in RCA]");
 #endif
     /* decrease precision if sl gets bigger than a multiple of e */
     sl += l;
@@ -549,7 +547,7 @@ TamelyRamifiedCase(KRASNER_t *data)
     while (r)
     {
       long gr;
-      GEN p1 = FpXQ_pow(pol_x(data->v), stoi(r), data->uplr, data->p);
+      GEN p1 = FpXQ_powu(pol_x(data->v), r, data->uplr, data->p);
       eis = gadd(Xe, ZX_Z_mul(p1, data->p)); /* Adding a ZX and a ZY (cste coefficient) */
       ct++;
       topx = get_topx(data, eis);
@@ -559,12 +557,8 @@ TamelyRamifiedCase(KRASNER_t *data)
       gel(vpl, ct) = eis;
       FieldData(data, &fdata, eis, topx);
       for (j = 1; j < ct; j++)
-      {
-        nb = IsIsomorphic(data, &fdata, gel(vpl, j));
-        if (nb)
-          pari_err(talker,
-                   "Oops, fields are isomorphic in TamelyRamifiedCase!\n");
-      }
+        if (IsIsomorphic(data, &fdata, gel(vpl, j)))
+          pari_err_BUG("TamelyRamifiedCase [isomorphic fields]");
       NbConjugateFields(data, &fdata);
       nb = fdata.cj;
       err_printf("Found %ld field(s)\n", nb);
@@ -584,7 +578,7 @@ TamelyRamifiedCase(KRASNER_t *data)
 
 #ifdef CHECK_EXTENSIONS
   if (!equaliu(data->nbext, cnt))
-    pari_err(talker,"Number of fields incorrect in TamelyRamifiedCase\n");
+    pari_err_BUG("TamelyRamifiedCase [incorrect #fields]");
 #endif
 
   return gerepilecopy(av, rep);
@@ -609,28 +603,21 @@ StructureOmega(KRASNER_t *data, GEN *pnbpol)
   GEN nbpol, O = cgetg(data->e + 1, t_VEC);
   long i;
 
-  nbpol = gen_1;
-  for (i = 0; i < data->e; i++)
+  /* i = 0 */
+  gel(O, 1) = mkvecsmall2(1, 0);
+  nbpol = mulii(data->qm1, powiu(data->q, data->c - 1));
+  for (i = 1; i < data->e; i++)
   {
     long v_start, zero = 0;
-    GEN nbcf;
-    if (i == 0)
-    {
-      v_start = 1;
-      nbcf = mulii(data->qm1, powiu(data->q, data->c - 1));
-    }
+    GEN nbcf, p1;
+    v_start = function_l(data->p, data->a, data->b, i);
+    p1 = powiu(data->q, data->c - v_start);
+    if (i == data->b)
+      nbcf = mulii(p1, data->qm1);
     else
     {
-      GEN p1;
-      v_start = function_l(data->p, data->a, data->b, i);
-      p1 = powiu(data->q, data->c - v_start);
-      if (i == data->b)
-        nbcf = mulii(p1, data->qm1);
-      else
-      {
-        nbcf = mulii(p1, data->q);
-        zero = 1;
-      }
+      nbcf = mulii(p1, data->q);
+      zero = 1;
     }
     gel(O, i+1) = mkvecsmall2(v_start, zero);
     nbpol = mulii(nbpol, nbcf);
@@ -710,7 +697,7 @@ WildlyRamifiedCase(KRASNER_t *data)
   Omega = StructureOmega(data, &nbpol);
   nbext = itos_or_0(data->nbext);
   if (!nbext || (nbext & ~LGBITS))
-    pari_err(talker,"too many extensions in padicfields");
+    pari_err_OVERFLOW("padicfields [too many extensions]");
 
   if (DEBUGLEVEL>0) {
     err_printf("There are %ld extensions to find and %Ps polynomials to consider\n", nbext, nbpol);
@@ -725,18 +712,16 @@ WildlyRamifiedCase(KRASNER_t *data)
   av2 = avma;
 
   while (fd < nbext)
-  {
-    /* The best thing seems to be to jump randomly among the polynomials... */
+  { /* Jump randomly among the polynomials : seems best... */
     rpl = RandomPol(data, Omega);
     if (DEBUGLEVEL>3) err_printf("considering polynomial %Ps\n", rpl);
-
 #ifdef CHECK_EXTENSIONS
     {
       GEN disc = poldisc0(rpl, 0);
       long e = data->e, f = data->f, j = data->j;
       disc = RgXQ_norm(disc, data->upl);
       if (Z_pval(disc, data->p) != f*(e+j-1))
-        pari_err(talker, "incorrect discriminant in WildlyRamifiedCase\n");
+        pari_err_BUG("WildlyRamifiedCase [incorrect discriminant]");
     }
 #endif
 
@@ -778,16 +763,16 @@ WildlyRamifiedCase(KRASNER_t *data)
 
 /* return the minimal polynomial (mod pr) of a generator of (F_p^f)^x with variable v */
 static GEN
-CycloPol(KRASNER_t *data)
+CycloPol(KRASNER_t *d)
 {
   GEN T, z;
   /* v - primroot(p) */
-  if (data->f == 1) return deg1pol_shallow(gen_1, Fp_neg(pgener_Fp(data->p), data->pr), data->v);
-
-  T = init_Fq(data->p, data->f, data->v);
-  z = gener_FpXQ(T, data->p, NULL);
-  z = ZpXQ_sqrtnlift(gen_1, data->qm1, z, T, data->p, data->r);
-  return FpX_red(ZXQ_charpoly(z, T, data->v), data->pr);
+  if (d->f == 1)
+    return deg1pol_shallow(gen_1, Fp_neg(pgener_Fp(d->p), d->pr), d->v);
+  T = init_Fq(d->p, d->f, d->v);
+  z = gener_FpXQ(T, d->p, NULL);
+  z = ZpXQ_sqrtnlift(scalarpol(gen_1,varn(T)), d->qm1, z, T, d->p, d->r);
+  return FpX_red(ZXQ_charpoly(z, T, d->v), d->pr);
 }
 
 /* return [ p^1, p^2, ..., p^c ] */
@@ -806,10 +791,11 @@ get_pk(KRASNER_t *data)
    where z is a root of upl defining an unramified extension of Q_p */
 /* See padicfields for the meaning of flag */
 static GEN
-GetRamifiedPol(GEN p, GEN efj, long v, long flag)
+GetRamifiedPol(GEN p, GEN efj, long flag)
 {
-  long e = efj[1], f = efj[2], j = efj[3];
-  GEN pols;
+  long e = efj[1], f = efj[2], j = efj[3], i, l;
+  const long v = 1;
+  GEN L, pols;
   KRASNER_t data;
   pari_sp av = avma;
 
@@ -841,40 +827,36 @@ GetRamifiedPol(GEN p, GEN efj, long v, long flag)
   data.roottable = NULL;
   if (j)
   {
-    GEN npol = powiu(data.q, e+1);
-    if (lgefint(data.p) == 3 && expi(npol) < 19)
+    if (lgefint(data.q) == 3)
     {
-      long i, l = itou(npol);
-      data.roottable = cgetg(l+1, t_VEC);
-      for (i = 1; i <= l; i++) gel(data.roottable, i) = NULL;
+      ulong npol = upowuu(data.q[2], e+1);
+      if (npol && npol < (1<<19)) data.roottable = const_vec(npol, NULL);
     }
     data.pk = get_pk(&data);
-    pols = WildlyRamifiedCase(&data);
+    L = WildlyRamifiedCase(&data);
   }
   else
-    pols = TamelyRamifiedCase(&data);
+    L = TamelyRamifiedCase(&data);
 
+  pols = cgetg_copy(L, &l);
   if (flag == 1)
   {
-    long i, l;
-    GEN p1 = cgetg_copy(pols, &l);
+    GEN E = utoipos(e), F = utoipos(f), D = utoi(f*(e+j-1));
     for (i = 1; i < l; i++)
-      gel(p1, i) = mkvec5(gcopy(gmael2(pols, i, 1)),
-                          utoipos(e),
-                          utoipos(f),
-                          utoi(f*(e+j-1)),
-                          gcopy(gmael2(pols, i, 2)));
-    pols = gerepileupto(av, p1);
+    {
+      GEN T = gel(L,i);
+      gel(pols, i) = mkvec5(ZX_copy(gel(T, 1)), E,F,D, icopy(gel(T, 2)));
+    }
   }
   else
   {
-    long i, l;
-    GEN p1 = cgetg_copy(pols, &l);
     for (i = 1; i < l; i++)
-      gel(p1, i) = gcopy(gmael2(pols, i, 1));
-    pols = gerepilecopy(av, p1);
+    {
+      GEN T = gel(L,i);
+      gel(pols, i) = ZX_copy(gel(T,1));
+    }
   }
-  return pols;
+  return gerepileupto(av, pols);
 }
 
 static GEN
@@ -891,7 +873,7 @@ possible_efj(GEN p, long m)
     for (pve = 1,ve = 1; ve <= v; ve++) { pve *= pp; nb += pve * ve; }
     nb = itos_or_0(muluu(nb, zv_sum(D)));
     if (!nb || is_bigint( mului(pve, sqru(v)) ) )
-      pari_err(talker,"too many ramification possibilities in padicfields");
+      pari_err_OVERFLOW("padicfields [too many ramification possibilities]");
   }
   nb += taum1; /* upper bound for the number of possible triples [e,f,j] */
 
@@ -924,10 +906,10 @@ possible_efj(GEN p, long m)
 static GEN
 pols_from_efj(pari_sp av, GEN EFJ, GEN p, long flag)
 {
-  long i, l, v = fetch_user_var("y");
+  long i, l;
   GEN L = cgetg_copy(EFJ, &l);
   if (l == 1) { avma = av; return flag == 2? gen_0: cgetg(1, t_VEC); }
-  for (i = 1; i < l; i++) gel(L,i) = GetRamifiedPol(p, gel(EFJ,i), v, flag);
+  for (i = 1; i < l; i++) gel(L,i) = GetRamifiedPol(p, gel(EFJ,i), flag);
   if (flag == 2) return gerepileuptoint(av, ZV_sum(L));
   return gerepilecopy(av, shallowconcat1(L));
 }
@@ -945,19 +927,21 @@ padicfields0(GEN p, GEN N, long flag)
   long m = 0, d = -1;
   GEN L;
 
-  if (typ(p) != t_INT) pari_err(arither1);
+  if (typ(p) != t_INT) pari_err_TYPE("padicfields",p);
   /* be nice to silly users */
-  if (!BPSW_psp(p)) pari_err(talker,"p must be a prime in padicfields");
+  if (!BPSW_psp(p)) pari_err_PRIME("padicfields",p);
   switch(typ(N))
   {
     case t_VEC:
-      if (lg(N) != 3 || typ(gel(N,2)) != t_INT) pari_err(typeer, "padicfields");
-      d = itos(gel(N,2));
+      if (lg(N) != 3) pari_err_TYPE("padicfields",N);
+      d = gtos(gel(N,2));
       N = gel(N,1); /* fall through */
     case t_INT:
       m = itos(N);
-      if (m <= 0) pari_err(talker,"non-positive degree in padicfields()");
+      if (m <= 0) pari_err_DOMAIN("padicfields", "degree", "<=", gen_0,N);
       break;
+    default:
+      pari_err_TYPE("padicfields",N);
   }
   if (d >= 0) return padicfields(p, m, d, flag);
   L = possible_efj(p, m);
